#' Choose a Folder Interactively
#'
#' Display an OS-native folder selection dialog under Mac OS X, Linux GTK+ or
#' Windows.
#'
#' @param default which folder to show initially
#' @param caption the caption on the selection dialog
#'
#' @details
#' Uses an Apple Script, Zenity or Windows Batch script to display an OS-native
#' folder selection dialog.
#'
#' For Apple Script, with \code{default = NA}, the initial folder selection
#' is determined by default behavior of the "choose folder" script. Otherwise,
#' paths are expanded with \link{path.expand}.
#'
#' For Windows Batch script the initial folder is always ignored.
#'
#' @return
#' A length one character vector, character NA if 'Cancel' was selected.
#'
if (Sys.info()['sysname'] == 'Darwin') {
  choose.dir = function(default = NA, caption = NA) {
    command = 'osascript'
    args = '-e "POSIX path of (choose folder{{prompt}}{{default}})"'

    if (!is.null(caption) && !is.na(caption) && nzchar(caption)) {
      prompt = sprintf(' with prompt \\"%s\\"', caption)
    } else {
      prompt = ''
    }
    args = sub('{{prompt}}', prompt, args, fixed = T)

    if (!is.null(default) && !is.na(default) && nzchar(default)) {
      default = sprintf(' default location \\"%s\\"', path.expand(default))
    } else {
      default = ''
    }
    args = sub('{{default}}', default, args, fixed = T)

    suppressWarnings({
      path = system2(command, args = args, stderr = TRUE)
    })
    if (!is.null(attr(path, 'status')) && attr(path, 'status')) {
      # user canceled
      path = NA
    } else {
      # cut any extra output lines, like "Class FIFinderSyncExtensionHost ..."
      path = tail(path, n=1)
    }

    return(path)
  }
} else if (Sys.info()['sysname'] == 'Linux') {
  choose.dir = function(default = NA, caption = NA) {
    command = 'zenity'
    args = '--file-selection --directory --title="Choose a folder"'

    suppressWarnings({
      path = system2(command, args = args, stderr = TRUE)
    })

    #Return NA if user hits cancel
    if (!is.null(attr(path, 'status')) && attr(path, 'status')) {
      # user canceled
      return(default)
    }

    #Error: Gtk-Message: GtkDialog mapped without a transient parent
    if(length(path) == 2){
      path = path[2]
    }

    return(path)
  }
} else if (Sys.info()['sysname'] == 'Windows') {
  # Use batch script to circumvent issue w/ `choose.dir`/`tcltk::tk_choose.dir`
  # window popping out unnoticed in the back of the current window
  choose.dir = function(default = NA, caption = NA) {
      command = file.path('utils','choose_dir.bat')
      args = if (is.na(caption)) '' else sprintf('"%s"', caption)
      suppressWarnings({
        path = system2(command, args = args, stdout = TRUE)
      })
      if (path == 'NONE') path = NA
      return(path)
  }
}

.cutDirPath = function(x,maxchar=30) {  # cut file path for display in ui
  if ( length(x) != 1 ) return("")
  if ( nchar(x) <= maxchar ) return(x)
  s <- strsplit(x,split="/",fixed=TRUE)[[1]]
  n <- length(s)
  if ( n == 1 | nchar(s[n])>maxchar-3 ) return(paste0("...",substring(x,nchar(x)-maxchar+4)))
  cumchar <- rev(cumsum(rev(nchar(s)+1)))
  return(paste0("[...]/",paste(s[c(cumchar[1:(n-1)]<=maxchar-3,TRUE)],collapse="/")))
}

.chosenDirId = function(inputId) sprintf('%s__chosen_dir', inputId)
.displayedDirId = function(inputId) sprintf('%s__displayed_dir', inputId)

#' Directory Selection Control
#'
#' Create a directory selection control to select a directory on the server
#'
#' @param inputId The \code{input} slot that will be used to access the value
#' @param label Display label for the control, or NULL for no label
#' @param value Initial value.  Paths are exapnded via \code{\link{path.expand}}.
#'
#' @details
#' This widget relies on \link{\code{choose.dir}} to present an interactive
#' dialog to users for selecting a directory on the local filesystem.  Therefore,
#' this widget is intended for shiny apps that are run locally - i.e. on the
#' same system that files/directories are to be accessed - and not from hosted
#' applications (e.g. from shinyapps.io).
#'
#' @return
#' A directory input control that can be added to a UI definition.
#'
#' @seealso
#' \link{updateDirectoryInput}, \link{readDirectoryInput}, \link[utils]{choose.dir}
directoryInput = function(inputId, label = "", value = NULL, btn.label="...") {
  if (!is.null(value) && !is.na(value)) {
    value = path.expand(value)
  }

  labelTag = if (nzchar(label))
    shiny:::`%AND%`(label, tags$label(`for`=inputId,label))
    else NULL

  tagList(
    singleton(
      tags$head(
        tags$script(src = 'js/directory_input_binding.js')
      )
    ),

    div(
      class = 'form-group shiny-input-container directory-input-container',
      labelTag,
      div(
        class = 'input-group',
        style = 'display: block;',
        tags$button(
          class = 'btn btn-default directory-input',
          style = 'display: inline-block; float: left;',
          id = inputId,
          icon('folder-o'),
          HTML('&nbsp;'),
          btn.label
        ),
        # tags$input(
        #   class = 'form-control directory-input-chosen-dir',
        #   style = 'display: table-cell;',
        #   id = .chosenDirId(inputId),
        #   value = value,
        #   type = 'text',
        #   readonly = 'readonly'
        # )
        tags$input(
          class = 'form-control directory-input-chosen-dir',
          style = 'display: none;',
          id = .chosenDirId(inputId),
          value = value,
          type = 'text'
        ),
        tags$input(
          class = 'form-control directory-input-displayed-dir',
          style = 'display: inline-block;',
          id = .displayedDirId(inputId),
          value = .cutDirPath(value),
          type = 'text',
          readonly = 'readonly'
        )
      )
    )
  )

}

#' Change the value of a directoryInput on the client
#'
#' @param session The \code{session} object passed to function given to \code{shinyServer}.
#' @param inputId The id of the input object.
#' @param value A directory path to set
#' @param ... Additional arguments passed to \link{\code{choose.dir}}.  Only used
#'    if \code{value} is \code{NULL}.
#'
#' @details
#' Sends a message to the client, telling it to change the value of the input
#' object.  For \code{directoryInput} objects, this changes the value displayed
#' in the text-field and triggers a client-side change event.  A directory
#' selection dialog is not displayed.
#'
updateDirectoryInput = function(session, inputId, value = NULL, ...) {
  if (is.null(value)) {
    value = choose.dir(...)
  }
  # session$sendInputMessage(inputId, list(chosen_dir = value))
  session$sendInputMessage(inputId, list(chosen_dir=value,
    displayed_dir=.cutDirPath(value)))
}

#' Read the value of a directoryInput
#'
#' @param session The \code{session} object passed to function given to \code{shinyServer}.
#' @param inputId The id of the input object
#'
#' @details
#' Reads the value of the text field associated with a \code{directoryInput}
#' object that stores the user selected directory path.
#'
readDirectoryInput = function(session, inputId) {
  session$input[[.chosenDirId(inputId)]]
}
