# ecovalAPI.R

library(ecoval)
library(stringr)
# required but avail via global.R: source("uiText.R")



outputLogger = list(
  errorPrefix = 'ERROR:',
  warningPrefix = 'WARN:',
  #' Function wrapper to capture error, if there is one, and all extra output
  #' messages (produced by error, warnings, messages, or stdout) to a temporary
  #' log file .
  #'
  #' Returns list with:
  #'    $value: function value or NULL if there was an error
  #'    $logStr: log string with all additional output as it would appear in the
  #'             shell
  #'    $err: error message or NULL if there was none
  fwrap = function(fun) {
    function(...) {
      # Note: using temp file and `sink` instead of just `withCallingHandlers`
      #       or `capture.output` to have both stdout and messages exactly
      #       in the same order as they would appear in the shell.
      fn = tempfile('ecovalui_calllog_')

      f = file(fn)
      sink(f, append=TRUE)
      sink(f, append=TRUE, type="message")

      err <- NULL
      val <- withCallingHandlers(
        tryCatch(
          fun(...),
          error = function(e) {
            err <<- conditionMessage(e)
            # put the error message also in the log
            message(paste(outputLogger$errorPrefix, err))
            NULL
          },
          finally = {
            sink(type="message")
            sink()
            close(f)
          }
        ),
        warning = function(w) {
          # in case of an error warnings will be the logged (otherwise they
          # would be masqueraded)
          message(paste(outputLogger$warningPrefix, conditionMessage(w)))
          invokeRestart("muffleWarning")
        }
      )

      logStr <- paste(readLines(fn), collapse = "\n")
      file.remove(fn)
      list(value=val, logStr=logStr, error=err)
    }
  },
  getValue = function(resWrap) resWrap$value,
  getLogStr = function(resWrap) resWrap$logStr,
  getError = function(resWrap) resWrap$error,
  hasError = function(resWrap) !(is.null(outputLogger$getError(resWrap))),
  hasLogError = function(resWrap) grepl(paste0("\n+",outputLogger$errorPrefix),
      outputLogger$getLogStr(resWrap))
)



.helpers = list(
  strtrimIfChar = function(x) {
    if (is.character(x)) str_trim(x) else x
  },
  strtrimArgsWrap = function(fun) {
    function(...) do.call(fun, lapply(list(...), .helpers$strtrimIfChar))
  },
  valueNotGiven = function(x = NA) {
    return (is.na(x) || is.character(x) && nchar(str_trim(x)) == 0)
  },
  fnHasExtension = function(fn) {
    .helpers$valueNotGiven(fn) || grepl("\\.\\w+", fn)
  },
  appendFnExtIfMissing = function(fn, ext=".txt") {
    if (!.helpers$fnHasExtension(fn)) {
        fn = paste0(fn, ext)
    }
    fn
  },
  .getFilePath = function(dir, basename, ext) {
    return(file.path(dir, .helpers$appendFnExtIfMissing(basename, ext)))
  },
  .getFilePathOrNA = function(dir, basename, ...) {
    ret = if (.helpers$valueNotGiven(dir) || .helpers$valueNotGiven(basename)) {
      NA
    } else {
      .helpers$.getFilePath(dir, basename, ...)
    }
    return(ret)
  }
)



AttributesModificationBuffer <- setRefClass("AttributesModificationBuffer",
  #' Modification buffer consists of ID (vector, in particluar single number
  #' or a string) and a list of modified attributes (properties) with their
  #' names (as keys) and their values.
  #'
  fields = list(
    idVec = "character", # default: character(0)
    modValuesList = "list"),
  methods = list(
    #' Clear buffer. If given use an identifier. Otherwise, use the default
    #' identifier.
    #'
    clear = function(idVec = character(0)) {
      .self$idVec <- idVec
      .self$modValuesList <- list()
    },
    #' Buffer attribute value getter
    #'
    getModValueOrDefault = function(propName, defaultPropValue) {
      # Attention: assume that IDs were checked before
      if (propName %in% names(.self$modValuesList))
          return(.self$modValuesList[[propName]])
      return(defaultPropValue)
    },
    #' Buffer attribute value setter
    #'
    setModValue = function(propName, propValue) {
      .self$modValuesList[[propName]] <- propValue
      return(propValue)
    },
    #' IDs equality
    #'
    isIdEqual = function(idVec) {
      selfId <- .self$idVec
      return(
        length(selfId) > 0 &&
        is.vector(idVec) &&
        (length(selfId) == length(idVec)) &&
        all(selfId == idVec))
    }
  )
)

AttributesModificationBufferPerObservation <- setRefClass(
  "AttributesModificationBufferPerObservation",
  contains = "AttributesModificationBuffer",
  methods = list(
    #' Compute buffer's ID. Buffer's ID is used to keep values only if the
    #' computed ID matches the buffer's one; \code{{get,set}ObservationValue}
    #' are the helper functions to achieve that.
    #'
    getBufferId = function(observationCodes, observationCode, ...) { # ... == propName
      # To change buffering behaviour (per prop, per obs, or per results set)
      # change how the buffer ID is computed.
      # Currently, character triplet: resId, observationCode,
      # observationCodeIdx, where resId is a single character vector of all
      # observation codes
      observationCodeIdx <- .self$getObservationCodeIdx(
        observationCodes,
        observationCode
      )
      resId <- paste(observationCodes, collapse=";")
      return(c(resId, observationCode, observationCodeIdx))
    },
    getObservationCodeIdx = function(observationCodes, observationCode) {
      observationCodeIdx <- which(observationCodes == observationCode,
        arr.ind = TRUE)
      if (length(observationCodeIdx) == 0) {
        stop(paste(uiText$observation.code, paste0("\"",observationCode,"\""),
          uiText$not.found))
      }
      return(observationCodeIdx)
    },
    #' Get buffered value of a property if buffer ID matches computed one,
    #' otherwise reset buffer and return corresponding property value from
    #' \code{res} list.
    #'
    getObservationValue = function(observationCodes, observationCode, propName, defaultPropValue) {
      bufId <- .self$getBufferId(observationCodes, observationCode, propName)
      if (!.self$isIdEqual(bufId)) .self$clear(idVec=bufId)
      returnValue <- .self$getModValueOrDefault(propName, defaultPropValue)
      return(returnValue)
    },
    #' Buffer new value of a property. Clear buffer first if buffer ID does not
    #' match the computed one.
    #'
    setObservationValue = function(observationCodes, observationCode, propName, propValue) {
      bufId <- .self$getBufferId(observationCodes, observationCode, propName)
      if (!.self$isIdEqual(bufId)) .self$clear(idVec=bufId)
      returnValue <- .self$setModValue(propName, propValue)
      return(returnValue)
    }
  )
)



################################
# lakes
################################
lakesAPI = list(
  #### Helpers
  getMorphology = .helpers$strtrimArgsWrap(
    function(dictionaryLanguage) {
      return(lake.morphol.2016.create(dictionaryLanguage))
    }
  ),
  dictionaryLanguages = list(
    choices = colnames(ecoval.dictionaries.default)[-c(1,2)],
    default = uiText$lang.code.default
  ),
  defaultFileExtensions = list(
      output.file.attributes = ".csv",
      output.file.assesments = ".csv",
      output.file.evaluation.hierarchies = ".pdf",
      output.file.evaluation.lines = ".pdf"
  ),
  .hierarchyLevelsValues = list(
    "E01" =
      c("E01.01",
        "E01.02",
        "E01.03",
        "E01.04",
        "E01.05",
        "E01.06",
        "E01.0601",
        "E01.0602",
        "E01.0603",
        "E01.07",
        "E01.08"),
    "E02" =
      c("E02.01",
        "E02.02",
        "E02.0201",
        "E02.0202",
        "E02.0203",
        "E02.0204",
        "E02.03",
        "E02.0301",
        "E02.0302",
        "E02.0303",
        "E02.0304",
        "E02.0305",
        "E02.04",
        "E02.0401",
        "E02.0402",
        "E02.0403",
        "E02.0404",
        "E02.0405",
        "E02.0406",
        "E02.0407",
        "E02.0408",
        "E02.0409",
        "E02.0410",
        "E02.0411",
        "E02.0412",
        "E02.0413",
        "E02.05",
        "E02.0501",
        "E02.0502",
        "E02.0503",
        "E02.0504",
        "E02.0505"),
    "B01" =
      c("B01.01",
        "B01.02",
        "B01.0201",
        "B01.0202",
        "B01.0203",
        "B01.03",
        "B01.0301",
        "B01.0302",
        "B01.0303",
        "B01.04",
        "B01.0401",
        "B01.0402",
        "B01.0403",
        "B01.0404",
        "B01.0405",
        "B01.05",
        "B01.0501",
        "B01.0502",
        "B01.0503",
        "B01.0504",
        "B01.0505",
        "B01.0506",
        "B01.0507",
        "B01.0508",
        "B01.06"),
    "B02" =
      c("B02.01",
        "B02.02",
        "B02.03",
        "B02.04"),
    "C01" =
      c("C01.01",
        "C01.02",
        "C01.03",
        "C01.0301",
        "C01.0302",
        "C01.0303",
        "C01.04",
        "C01.05",
        "C01.0501",
        "C01.0502",
        "C01.0503",
        "C01.0504",
        "C01.06",
        "C01.0601",
        "C01.0602",
        "C01.0603",
        "C01.0604",
        "C01.0605",
        "C01.07",
        "C01.0701",
        "C01.0702",
        "C01.0703",
        "C01.0704"),
    "C02" =
      c("C02.01",
        "C02.02",
        "C02.03",
        "C02.0301",
        "C02.0302",
        "C02.0303",
        "C02.0304",
        "C02.0305",
        "C02.0306",
        "C02.04",
        "C02.0401",
        "C02.0402",
        "C02.0403",
        "C02.0404",
        "C02.0405",
        "C02.0406",
        "C02.05",
        "C02.0501"),
    "C03" =
      c("C03.01",
        "C03.02",
        "C03.03",
        "C03.04",
        "C03.05",
        "C03.06",
        "C03.07",
        "C03.08",
        "C03.09",
        "C03.10",
        "C03.11"),
    "C04" =
      c("C04.01",
        "C04.02",
        "C04.0201",
        "C04.0202",
        "C04.0203",
        "C04.0204",
        "C04.04",
        "C04.03",
        "C04.0301",
        "C04.0302",
        "C04.0303",
        "C04.0304"),
    "C05" =
      c("C05.01",
        "C05.02",
        "C05.03",
        "C05.04",
        "C05.05",
        "C05.06",
        "C05.07"),
    "C06" =
      c("C06.01",
        "C06.02",
        "C06.03",
        "C06.04"),
    "D01" =
      c("D01.01",
        "D01.02",
        "D01.0201",
        "D01.0202",
        "D01.0203",
        "D01.0204",
        "D01.03",
        "D01.0301",
        "D01.0302",
        "D01.0303",
        "D01.0304",
        "D01.04",
        "D01.0401",
        "D01.0402",
        "D01.0403",
        "D01.0404",
        "D01.0405",
        "D01.0406",
        "D01.05",
        "D01.0501",
        "D01.0502",
        "D01.0503",
        "D01.0504"),
    "D02" =
      c("D02.01",
        "D02.02",
        "D02.03",
        "D02.04")),
  getHierarchyLevelValuesForHierarchyLevel = function(attributes, hierarchyLevel) {
    if (!(hierarchyLevel %in% names(lakesAPI$.hierarchyLevelsValues)))
      return(NULL)
    return(lakesAPI$.hierarchyLevelsValues[[hierarchyLevel]])
  },
  getEvaluation = function(morphology, attributes) {
    return(evaluate(morphology, attributes))
  },
  getSegmentNames = function(attributes) {
    return(rownames(attributes))
  },
  getHierarchyLevels = function(attributes) {
    colNames <- colnames(attributes)
    # Note: subset of all hierarchy lvls that perserves order of attributes cols
    hierarchyLevels <- colNames[colNames %in% names(lakesAPI$.hierarchyLevelsValues)]
    # # OPT: consider as a hierarchy level any column name in a format "Xnm" where
    # # X is a letter, and n,m are ditigs
    # hierarchyLevels <- colNames[grep("^[A-Z][[:digit:]]{2}$", colNames)]
    return(hierarchyLevels)
  },
  modBuffer = AttributesModificationBufferPerObservation$new(),
  getHierarchyLevelValuesForSegment = .helpers$strtrimArgsWrap(
    function(attributes, segmentName) {
      # return(attributes[observationCodeOrIdx, lakesAPI$getHierarchyLevels(attributes)])
      segNames <- lakesAPI$getSegmentNames(attributes)
      hierarchyLevels <- lakesAPI$getHierarchyLevels(attributes)
      return(lakesAPI$modBuffer$getObservationValue(
        segNames,
        segmentName,
        "hierarchy.levels",
        attributes[segmentName, hierarchyLevels]
      ))
    }
  ),
  getHierarchyLevelValueForSegment = .helpers$strtrimArgsWrap(
    function(attributes, segmentName, hierarchyLevel) {
      hierarchyLevels <- lakesAPI$getHierarchyLevels(attributes)
      if (!(hierarchyLevel %in% hierarchyLevels)) return(character(0))
      segHierarchyLevels <- lakesAPI$getHierarchyLevelValuesForSegment(
        attributes,
        segmentName
      )
      return(segHierarchyLevels[hierarchyLevel])
    }
  ),
  setHierarchyLevelValueForSegment = .helpers$strtrimArgsWrap(
    function(attributes, segmentName, hierarchyLevel, hierarchyLevelValue) {
      hierarchyLevels <- lakesAPI$getHierarchyLevels(attributes)
      # # DEBUG
      # print("setHierarchyLevelValueForSegment")
      # print(hierarchyLevels)
      # print(hierarchyLevel %in% hierarchyLevels)
      # print(lakesAPI$.hierarchyLevelsValues[[hierarchyLevel]])
      # print(hierarchyLevelValue %in% lakesAPI$.hierarchyLevelsValues[[hierarchyLevel]])
      if (
        !(hierarchyLevel %in% hierarchyLevels) ||
        !(hierarchyLevelValue %in% lakesAPI$.hierarchyLevelsValues[[hierarchyLevel]])
      ) return(character(0))
      segNames <- lakesAPI$getSegmentNames(attributes)
      segHierarchyLevels <- lakesAPI$getHierarchyLevelValuesForSegment(
        attributes,
        segmentName
      )
      segHierarchyLevels[hierarchyLevel] <- hierarchyLevelValue
      setValue <- lakesAPI$modBuffer$setObservationValue(
        segNames,
        segmentName,
        "hierarchy.levels",
        segHierarchyLevels
      )
      # # DEBUG
      # print(segmentName)
      # print(hierarchyLevel)
      # print(hierarchyLevelValue)
      # print(segHierarchyLevels)
      # print(setValue)
      return(setValue)
    }
  ),
  .getAttributesCol = .helpers$strtrimArgsWrap(
    function(attributes, colName) {
      # Note: works w/ data.frame (keeps single columns indexable by row) but
      #       not w/ matrix; `attributes[,colName]` works with both types but
      #       in case of indexing w/ a single col it always needs transpose and,
      #       for data.frame, setting back row names
      return(attributes[colName])
    }
  ),
  getSegmentsRange = function(attributes, startColName, endColName) {
    return(c(
      floor(min(lakesAPI$.getAttributesCol(attributes, startColName))),
      ceiling(max(lakesAPI$.getAttributesCol(attributes, endColName)))
    ))
  },
  getIndicesWithinRange = function(attributes, startColName, endColName, range) {
    # not relying on the fact that attributes are sorted by segments start & end
    starts = lakesAPI$.getAttributesCol(attributes, startColName)
    ends = lakesAPI$.getAttributesCol(attributes, endColName)
    return(which(starts >= range[1] & ends <= range[2]))
  },
  ## Helpers for processing or plotting of numerical results
  .writePdfWrapper = function(filePath, width, height, plotFun, ...) {
      pdf(filePath, height=as.numeric(height),width=as.numeric(width))
      plotFun(...)
      dev.off()
  },
  .evaluationCut = function(range, startColName, endColName, dictionaryLanguage, attributes) {
      # empty values to full range
      if (all(is.null(range)) || all(range == c(0,0)))
        range = lakesAPI$getSegmentsRange(attributes, startColName, endColName)
      selectedRows <- lakesAPI$getIndicesWithinRange(attributes, startColName, endColName, range)
      attrib.cut <- attributes[selectedRows,]
      morphology <- lakesAPI$getMorphology(dictionaryLanguage)
      evaluation.cut <- lakesAPI$getEvaluation(morphology, attrib.cut)
      return(list(
        attributes = attrib.cut,
        evaluation = evaluation.cut,
        morphology = morphology
      ))
  },
  #### ecoval API wrappers with logged output
  plotLakeMorphology = outputLogger$fwrap(
    function(dictionaryLanguage) {
      par(mar=c(0,0,0,0))
      plot(lakesAPI$getMorphology(dictionaryLanguage), main=dictionaryLanguage,
        two.lines=TRUE, cex.nodes=0.4, cex.attrib=0.8)
    }
  ),
  plotSingleHierarchy = outputLogger$fwrap(
    function(segmentName, dictionaryLanguage, attributes, with.attrib=FALSE) {
      # observationCodeIdx <- lakesAPI$.getObservationCodeIdx(
      #   attributes,
      #   segmentName)
      # if (length(observationCodeIdx) == 0) {
      #   stop(paste(
      #     uiText$segment,
      #     paste0("\"",segmentName,"\""),
      #     uiText$not.found))
      # }
      morphology <- lakesAPI$getMorphology(dictionaryLanguage)
      segHierarchyLevels <- lakesAPI$getHierarchyLevelValuesForSegment(
        attributes,
        segmentName
      )
      # for the evaluation transform the segment's hierarchy levels vec into the
      # data frame format of attributes
      segHierarchyLevelsDf <- data.frame(as.list(segHierarchyLevels))
      colnames(segHierarchyLevelsDf) <- names(segHierarchyLevels)
      rownames(segHierarchyLevelsDf) <- segmentName
      # # DEBUG
      # print("plotSingleHierarchy")
      # print(segmentName)
      # print(segHierarchyLevels)
      # print(segHierarchyLevelsDf)
      evaluationSingle <- lakesAPI$getEvaluation(morphology, segHierarchyLevelsDf)
      par(mar = c(0,0,0,0))
      plot(
        morphology,
        evaluationSingle,
        two.lines = TRUE,
        with.attrib = with.attrib,
        main = segmentName,
        cex.nodes = 0.8)
    }
  ),
  readInputFiles = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(inputDir, start, end, level, measureid=NA, comments=NA) {

      # Clean inputs
      # note: character inputs trimmed by wrapper
      # strip trialing comma from `comments` if there is any
      comments <- sub(",$", "", comments)

      # Read attributes files
      attrib <- lake.morphol.2016.read.attrib(
        directory = inputDir,
        col.names = c(start, end, level, measureid, comments)
      )

    # OPT: report missing data
    #   comments.cols <- if (is.na(comments)) "" else strsplit(comments,",")[[1]]
    #   ind.with.na <- which(apply(attrib[,colnames(attrib)!=comments.cols],1,anyNA));
    #   print(paste(length(ind.with.na),"Zeilen mit fehlenden Daten von",nrow(attrib)));
    #   if ( length(ind.with.na)>0 ) {
    #     for(i in ind.with.na) {
    #       print(paste(rownames(attrib)[i],
    #                   paste(colnames(attrib)[is.na(attrib[i,colnames(attrib)!=comments.cols])],collapse=","),sep=" : "))
    #     }
    #   }

      # Read-in CSV files as a single table
      return(attrib)
    }
  )),
  .writeTableWrapper = .helpers$strtrimArgsWrap(
    function(obj, filePath, separator) {
      write.table(obj, filePath, col.names=NA, row.names=TRUE, sep=separator, na="")
    }
  ),
  writeAttributesFile = outputLogger$fwrap(
    function(filePath, separator, dictionaryLanguage, attributes) {
      lakesAPI$.writeTableWrapper(attributes, filePath, separator)
      return(attributes)
    }
  ),
  writeBewertungenFile = outputLogger$fwrap(
    function(filePath, separator, dictionaryLanguage, attributes) {
      evaluation <- lakesAPI$getEvaluation(lakesAPI$getMorphology(dictionaryLanguage), attributes)
      lakesAPI$.writeTableWrapper(evaluation, filePath, separator)
      return(evaluation)
    }
  ),
  writeHierarchienFullFile = outputLogger$fwrap(
    function(filePath, width, height, range, startColName, endColName, dictionaryLanguage, attributes) {
      # # empty values to full range
      # if (all(is.null(range)) || all(range == c(0,0)))
      #   range = lakesAPI$getSegmentsRange(attributes, startColName, endColName)
      # selectedRows <- lakesAPI$getIndicesWithinRange(attributes, startColName, endColName, range)
      # attrib.cut = attributes[selectedRows,]
      # morphology <- lakesAPI$getMorphology(dictionaryLanguage)
      # evaluation.cut <- lakesAPI$getEvaluation(morphology, attrib.cut)
      cutted <- lakesAPI$.evaluationCut(range, startColName, endColName, dictionaryLanguage, attributes)
      lakesAPI$.writePdfWrapper(filePath, width, height,
        plot, cutted$morphology, cutted$evaluation,
        two.lines=TRUE, main=rownames(cutted$evaluation), cex.nodes=0.8)
      return(cutted)
    }
  ),
  writeBewertungsLinienFile = outputLogger$fwrap(
    function(filePath, width, height, measureIdColName, range, startColName, endColName, dictionaryLanguage, attributes) {
      cutted <- lakesAPI$.evaluationCut(range, startColName, endColName, dictionaryLanguage, attributes)
      lakesAPI$.writePdfWrapper(filePath, width, height,
        lake.morphol.2016.plot.val.spatial, cutted$evaluation,
        main=measureIdColName, lwd=3);
      return(cutted)
    }
  )
)



################################
# macrophytes
################################
macrophytesAPI = list(
  #### Helpers
  dictionaryLanguages = list(
    choices = colnames(ecoval.dictionaries.default)[-c(1,2)],
    default = uiText$lang.code.default
  ),
  samplingProtocols = list(
    choices = c("v2009", "v2018"),
    default = "v2018"
  ),
  defaultFileExtensions = list(
    input.file.site = ".txt",
    input.file.species = ".txt",
    input.file.typeplaus = ".txt",
    output.file.numeric = ".txt",
    output.file.doc = ".pdf",
    output.file.taxa.used = ".txt",
    output.file.taxa.removed = ".txt",
    output.file.quality.checks = ".txt"
  ),
  defaultDataSeparator = "\t",
  defaultSampleSize = 10000,
  .separatorClean = function(sep) {
    # separator default "\t" for when trimmed to ""
    # Note: "" means whitespaces in `{read,write}.table`, but it doesn't with
    # `msk.macrophytes.2017.read.compile.evaluate`
    if (nchar(sep) == 0) sep = macrophytesAPI$defaultDataSeparator
    return(sep)
  },
  getObservationCodes = function(res) {
    if (is.null(res) ) return(NULL)
    if (!is.list(res) || !all(c("language", "dictionaries", "data.site") %in% names(res)))
        stop("Invalid results argument")
    dict <- ecoval.dict(res$language,res$dictionaries)
    observationCodesColname <- ecoval.translate("A_macrophytes_site_sampleid",dict)
    if (!(observationCodesColname %in% colnames(res$data.site))) {
        # Happens e.g. if chosen input lang does not match the input files lang:
        warning(paste(uiText$column.name, paste0("\"",observationCodesColname,"\""), uiText$not.found))
        return(NULL)
    }
    return(res$data.site[,observationCodesColname])
  },
  .getObservationCodeIdx = .helpers$strtrimArgsWrap(function(res, observationCode) {
      observationCodes <- macrophytesAPI$getObservationCodes(res)
      observationCodeIdx <- which(observationCodes == observationCode, arr.ind = TRUE)
      if (length(observationCodeIdx) == 0) {
        stop(paste(uiText$observation.code, paste0("\"",observationCode,"\""), uiText$not.found))
      }
      return(observationCodeIdx)
  }),
  getAttributeTypes = function(res) {
    colnames(res$attrib.types)
  },
  # TODO: re-factor to use AttributesModificationBufferPerObservation class
  #' Modification buffer consists of ID (vector, in particluar single number or
  #' a string) and a list of modified properties names (keys) and their values.
  #'
  .modBuffer = list(idVec = NA, modValuesList = list()),
  modBuffer.get = function() {
    macrophytesAPI$.modBuffer
  },
  modBuffer.set = function(modBuffer) {
    macrophytesAPI$.modBuffer <<- modBuffer
  },
  #' Initialize a new modBuffer. If given use given identifier. Otherwise, a
  #' default empty identifier (NA) is used.
  #'
  modBuffer.new = function(idVec = NA) {
    macrophytesAPI$modBuffer.set(list(idVec = idVec, modValuesList = list()))
  },
  ## Buffer properties getters and setters
  modBuffer.getIdVec = function() macrophytesAPI$modBuffer.get()$idVec,
  modBuffer.getModValuesList = function() macrophytesAPI$modBuffer.get()$modValuesList,
  modBuffer.getModValueOrDefault = function(propName, defaultPropValue) {
    modValuesList = macrophytesAPI$modBuffer.getModValuesList()
    # Attention: assume that IDs were checked before
    if (propName %in% names(modValuesList))
        return(modValuesList[[propName]])
    return(defaultPropValue)
  },
  modBuffer.setModValue = function(propName, propValue) {
    mb <- macrophytesAPI$modBuffer.get()
    mb$modValuesList[[propName]] <- propValue
    macrophytesAPI$modBuffer.set(mb)
    return(propValue)
  },
  #' ID equality
  modBuffer.isIdEqual = function(idVec) {
    selfId = macrophytesAPI$modBuffer.getIdVec()
    return( !is.na(selfId) && is.vector(idVec) && length(selfId)==length(idVec)
      && all(selfId==idVec) )
  },

  #' Compute buffer's ID. Buffer's ID is used to keep values only if the
  #' computed ID matches the buffer's one. \code{.{get,set}ObservationValue} are
  #' the helper functions to achieve that.
  #'
  .getBufferId = function(res, observationCode, propName) {
    # To change buffering behaviour (per prop, per obs, or per results set)
    # change how the buffer ID is computed.
    # Currently, character triplet: resId, observationCode,
    # observationCodeIdx,where resId is a single character vector of all
    # observation codes
    observationCodeIdx <- macrophytesAPI$.getObservationCodeIdx(res,
      observationCode)
    resId <- paste(macrophytesAPI$getObservationCodes(res), collapse=";")
    c(resId, observationCode, observationCodeIdx)
  },
  #' Get buffered value of a property if buffer ID matches compouted one,
  #' otherwise reset buffer and return value from res.
  #'
  .getObservationValue = function(res, observationCode, propName) {
    bufId <- macrophytesAPI$.getBufferId(res, observationCode, propName)
    # print(".getObservationValue BEGIN")
    # print("  buf pre")
    # print(macrophytesAPI$modBuffer.getIdVec())
    # print(macrophytesAPI$modBuffer.getModValuesList())
    if (!macrophytesAPI$modBuffer.isIdEqual(bufId))
      macrophytesAPI$modBuffer.new(bufId)
    observationCodeIdx <- macrophytesAPI$.getObservationCodeIdx(res,
      observationCode)
    defaultPropDf <- res[[propName]]
    # using as.matrix for vector properties to also work with [i,] indexing
    if (is.vector(defaultPropDf)) defaultPropDf <- as.matrix(defaultPropDf)
    defaultPropValue <- defaultPropDf[observationCodeIdx,]
    # print("  def val")
    # print(defaultPropValue)
    returnValue <- macrophytesAPI$modBuffer.getModValueOrDefault(propName,
      defaultPropValue)
    # print("  buf post")
    # print(macrophytesAPI$modBuffer.getIdVec())
    # print(macrophytesAPI$modBuffer.getModValuesList())
    # print("  ret val")
    # print(returnValue)
    # print(".getObservationValue END")
    returnValue
  },
  #' Buffer new value of a property. Clear buffer first if buffer ID does not
  #' match the computed one.
  #'
  .setObservationValue = function(res, observationCode, propName, propValue) {
    # print(".setObservationValue BEGIN")
    # print(propName)
    # print(propValue)
    bufId <- macrophytesAPI$.getBufferId(res, observationCode, propName)
    # print("  buf pre")
    # print(macrophytesAPI$modBuffer.getIdVec())
    # print(macrophytesAPI$modBuffer.getModValuesList())
    if (!macrophytesAPI$modBuffer.isIdEqual(bufId))
      macrophytesAPI$modBuffer.new(bufId)
    returnValue <- macrophytesAPI$modBuffer.setModValue(propName, propValue)
    # print("  buf post")
    # print(macrophytesAPI$modBuffer.getIdVec())
    # print(macrophytesAPI$modBuffer.getModValuesList())
    # print("  buf val")
    # print(macrophytesAPI$modBuffer.getModValuesList()[[propName]])
    # print("  ret val")
    # print(returnValue)
    # print(".setObservationValue END")
    returnValue
  },
  ## Helpers (getters/setters/validators) for specific properties
  .getObservationAttributeTypes = function(res, observationCode) {
    macrophytesAPI$.getObservationValue(res, observationCode, "attrib.types")
  },
  .validateAttributeType = function(res, attributeType) {
    if (!(attributeType %in% macrophytesAPI$getAttributeTypes(res))) {
      stop(paste(uiText$attribute, paste0("\"",attributeType,"\""), uiText$not.found))
    }
  },
  getObservationAttributeTypeValue = .helpers$strtrimArgsWrap(function(res,
    observationCode, attributeType)
  {
    #macrophytesAPI$.validateAttributeType(res, attributeType)
    # FIXME: validate if possible or re-factor if decided not to validate
    if (!(attributeType %in% macrophytesAPI$getAttributeTypes(res))) return(NULL)
    attributeTypeValue <- macrophytesAPI$.getObservationAttributeTypes(res,
        observationCode)[attributeType]
    if (!is.null(attributeTypeValue))
        attributeTypeValue <- as.numeric(attributeTypeValue)
    attributeTypeValue
  }),
  setObservationAttributeTypeValue = .helpers$strtrimArgsWrap(function(res,
    observationCode, attributeType, attributeTypeValue)
  {
    #macrophytesAPI$.validateAttributeType(res, attributeType)
    # FIXME: validate if possible or re-factor if decided not to validate
    if (!(attributeType %in% macrophytesAPI$getAttributeTypes(res))) return(NULL)
    obs.attrib.types <- macrophytesAPI$.getObservationAttributeTypes(res,
      observationCode)
    if (!is.null(attributeTypeValue))
        attributeTypeValue <- as.numeric(attributeTypeValue)
    obs.attrib.types[attributeType] <- attributeTypeValue
    macrophytesAPI$.setObservationValue(res, observationCode, "attrib.types",
      obs.attrib.types)
  }),
  getRiverTypes = function(res) {
    # no need to check `hasRiverTypes` - `colnames(NULL)` gives `NULL`
    colnames(res$types.val.probs)
  },
  .validateRiverType = function(res, riverType) {
    if (!(riverType %in% macrophytesAPI$getRiverTypes(res))) {
      stop(paste(uiText$river.type, paste0("\"",riverType,"\""), uiText$not.found))
    }
  },
  .riverTypeArgName = function(final=FALSE) {
    paste0("types.val.", if (final) "final" else "obs")
  },
  getObservationRiverType = function(res, observationCode, ...) {
    riverTypeArgName = macrophytesAPI$.riverTypeArgName(...)
    if ( riverTypeArgName %in% names(res) ) {
      as.vector(macrophytesAPI$.getObservationValue(res, observationCode,
        riverTypeArgName))
    } else NULL
  },
  setObservationRiverType = .helpers$strtrimArgsWrap(function(res,
    observationCode, riverType, ...)
  {
    riverTypeArgName = macrophytesAPI$.riverTypeArgName(...)
    if (!(riverTypeArgName %in% names(res))) return(NA)
    #macrophytesAPI$.validateRiverType(res, riverType)
    # FIXME: validate if possible or re-factor if decided not to validate
    if (!(riverType %in% macrophytesAPI$getRiverTypes(res))) return(NA)
    macrophytesAPI$.setObservationValue(res, observationCode, riverTypeArgName,
      as.character(riverType))
  }),
  getObservationRiverTypeFinal = function(...) {
    macrophytesAPI$getObservationRiverType(..., final=TRUE)
  },
  setObservationRiverTypeFinal = .helpers$strtrimArgsWrap(function(...)
  {
    macrophytesAPI$setObservationRiverType(..., final=TRUE)
  }),
  getAttributeSpecies = function(res) {
    #ret = grep("^(n|dg|heip|moos)\\.", colnames(res$attrib.species), value=TRUE)
    ret <- NULL
    if ( !anyNA(match(c("language","dictionaries","valfun.macrophytes"),names(res))) ) {
      dict <- ecoval.dict(res$language,res$dictionaries)
      att.names <- utility.get.attrib.names(res$valfun.macrophytes)
      ret <- att.names[att.names!=ecoval.translate("A_macrophytes_rivertype_class",dict)]
    }
    return(ret)
  },
  .getObservationAttributeSpecies = function(res, observationCode) {
    macrophytesAPI$.getObservationValue(res, observationCode, "attrib.species")
  },
  .validateAttributeSpecies = function(res, attributeSpecies) {
    if (!(attributeSpecies %in% macrophytesAPI$getAttributeSpecies(res))) {
      stop(paste(uiText$attribute, paste0("\"",attributeSpecies,"\""), uiText$not.found))
    }
  },
  getObservationAttributeSpeciesValue = .helpers$strtrimArgsWrap(function(res,
    observationCode, attributeSpecies)
  {
    #macrophytesAPI$.validateAttributeSpecies(res, attributeSpecies)
    # FIXME: validate if possible or re-factor if decided not to validate
    if (!(attributeSpecies %in% macrophytesAPI$getAttributeSpecies(res))) return(NULL)
    attributeSpeciesValue <- macrophytesAPI$.getObservationAttributeSpecies(res,
        observationCode)[attributeSpecies]
    if (!is.null(attributeSpeciesValue))
        attributeSpeciesValue <- as.numeric(attributeSpeciesValue)
    attributeSpeciesValue
  }),
  setObservationAttributeSpeciesValue = .helpers$strtrimArgsWrap(function(res,
    observationCode, attributeSpecies, attributeSpeciesValue)
  {
    #macrophytesAPI$.validateAttributeSpecies(res, attributeSpecies)
    # FIXME: validate if possible or re-factor if decided not to validate
    if (!(attributeSpecies %in% macrophytesAPI$getAttributeSpecies(res))) return(NULL)
    obs.attrib.species <- macrophytesAPI$.getObservationAttributeSpecies(res,
      observationCode)
    if (!is.null(attributeSpeciesValue))
        attributeSpeciesValue <- as.numeric(attributeSpeciesValue)
    obs.attrib.species[attributeSpecies] <- attributeSpeciesValue
    macrophytesAPI$.setObservationValue(res, observationCode, "attrib.species",
      obs.attrib.species)
  }),
  ##  Helpers for processing or plotting of numerical results
  .hasRiverType = function(res) {
    return("types.table" %in% names(res))
  },
  hasTypification = function(...) macrophytesAPI$.hasRiverType(...),
  hasAssesment = function(res) {
    return("val" %in% names(res))
  },
  .hasRiverTypePlausible = function(res) {
    return("types.plaus" %in% names(res) && length(res$types.plaus) > 0)
  },
  hasAssesmentPlausible = function(...) {
    macrophytesAPI$hasAssesment(...) && macrophytesAPI$.hasRiverTypePlausible(...)
  },
  .calcTypesOfObservation = function(observationCode, res,
      sampleSize = macrophytesAPI$defaultSampleSize,
      dictionaryLanguage = macrophytesAPI$dictionaryLanguages$default)
  {
      # requires river types to be pre-computed already
      if (!macrophytesAPI$hasTypification(res))
        stop(uiText$compute.typification.first)
      attrib.types.i <- macrophytesAPI$.getObservationAttributeTypes(res,
        observationCode)
      res.calc.types <- msk.macrophytes.2017.calc.types(attrib.types.i,
        sampsize = sampleSize, language = dictionaryLanguage)
      # copy manually lang and dicts, otherwise can't use plotting funs
      res.calc.types$language     <- res$language
      res.calc.types$dictionaries <- res$dictionaries
      return(res.calc.types)
  },
  .calcValfunOfObservation = function(observationCode, res, rivertype) {
      dict <- ecoval.dict(res$language,res$dictionaries)
      attrib <- data.frame(
        Flusstyp=rivertype,
        macrophytesAPI$.getObservationAttributeSpecies(res, observationCode))
      colnames(attrib)[1] <- ecoval.translate("A_macrophytes_rivertype_class",dict)
      res.valfun <- evaluate(res$valfun.macrophytes, attrib)
      return(res.valfun)
  },
  .plotAssesment = function(res, rivertype, val, with.attrib=FALSE, ...) {
    dict <- ecoval.dict(res$language,res$dictionaries)
    moss <- c(
      ecoval.translate("L_macrophytes_rivertype_class_smallbryophyte",dict),
      ecoval.translate("L_macrophytes_rivertype_class_mediumbryophyte",dict),
      ecoval.translate("L_macrophytes_rivertype_class_largebryophyte",dict),
      ecoval.translate("L_macrophytes_rivertype_class_verylargebryophyte",dict))
    top.node <- ecoval.translate("N_macrophytes_goodstate",dict)
    if ( rivertype %in% moss )
      top.node <- paste(ecoval.translate("N_macrophytes_goodstate",dict),
        ecoval.translate("N_macrophytes_bryophytesriver",dict))
    par(mar=c(0,0,0,0))
    plot(res$valfun.macrophytes, val, nodes=top.node, two.lines=TRUE,
      with.attrib=with.attrib,cex.nodes=ifelse(rivertype %in% moss,1,0.8))
  },
  #### ecoval API wrappers with logged output
  readInputFilesEvaluateAndWrite = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(dictionaryLanguage, inputDirData, inputDirImg, inputSeparator, inFileSites, inFileSpecies = NA, inFilePlausibleType = NA,
             samplingProtocol = macrophytesAPI$samplingProtocol$default, sampleSize = macrophytesAPI$defaultSampleSize,
             outputDir = NA, outputSeparator = "", outFileNumeric = NA, outFileDoc = NA, outFileTaxaUsed = NA, outFileTaxaRemoved = NA, outFileQualityChecks = NA) {
      inputSeparator = macrophytesAPI$.separatorClean(inputSeparator)
      outputSeparator = macrophytesAPI$.separatorClean(outputSeparator)
      # prepare files paths
      input.file.site = .helpers$.getFilePath(inputDirData, inFileSites,
        macrophytesAPI$defaultFileExtensions$input.file.site)
      input.file.species = .helpers$.getFilePathOrNA(inputDirData, inFileSpecies,
        macrophytesAPI$defaultFileExtensions$input.file.species)
      input.file.typeplaus = .helpers$.getFilePathOrNA(inputDirData, inFilePlausibleType,
        macrophytesAPI$defaultFileExtensions$input.file.typeplaus)
      output.file.numeric = .helpers$.getFilePathOrNA(outputDir, outFileNumeric,
        macrophytesAPI$defaultFileExtensions$output.file.numeric)
      output.file.doc = .helpers$.getFilePathOrNA(outputDir, outFileDoc,
        macrophytesAPI$defaultFileExtensions$output.file.doc)
      output.file.taxa.used = .helpers$.getFilePathOrNA(outputDir, outFileTaxaUsed,
        macrophytesAPI$defaultFileExtensions$output.file.taxa.used)
      output.file.taxa.removed = .helpers$.getFilePathOrNA(outputDir, outFileTaxaRemoved,
        macrophytesAPI$defaultFileExtensions$output.file.taxa.removed)
      output.file.quality.checks = .helpers$.getFilePathOrNA(outputDir, outFileQualityChecks,
        macrophytesAPI$defaultFileExtensions$output.file.quality.checks)
      # translate to the wrapped func args
      args <- list(
        file.site = input.file.site,
        pic.folder = inputDirImg,
        file.species = input.file.species,
        file.typeplaus = input.file.typeplaus,
        sep.in = inputSeparator,
        sep.out = outputSeparator,
        sampling.protocol = samplingProtocol,
        sampsize = sampleSize,
        file.check.msg = output.file.quality.checks,
        language = dictionaryLanguage,
        file.res = output.file.numeric,
        file.doc = output.file.doc,
        file.taxa.used = output.file.taxa.used,
        file.taxa.removed = output.file.taxa.removed
      )
      # revert to ecoval pkg defaults if arg value was not given; cf. pic.folder arg
      for (arg.name in names(args)) {
        if ( .helpers$valueNotGiven(args[[arg.name]]) ) args[[arg.name]] <- NULL
      }
      # read, opt. evaluate, and opt. write
      # Note: if sampleSize == 0, only write diagnostics w/o computing river
      #       types and w/o evaluation
      # Note: if outputDir or all ouput files are "NA", no output file will be
      #       written
      # cat diagnostic msgs - will be intercepted by the outputLogger
      cat(paste0("START: ", Sys.time(), "\n"))
      res <- do.call(msk.macrophytes.2017.read.compile.evaluate, args)
      dict <- ecoval.dict(res$language,res$dictionaries)
      # write error message if there were input errors, but do not `stop()` s.t.
      # checks file is linked and available
      if ( !is.null(res$site.quality) ) {
        if ( any(res$site.quality[,ecoval.translate("R_macrophytes_error_type",dict)]==ecoval.translate("R_macrophytes_error_error",dict)) )
          cat(paste(outputLogger$errorPrefix, paste0(uiText$check.fail.sites.data,":"),
            uiText$see, paste0(uiText$quality.checks,"\n")))
      }
      if ( !is.null(res$species.quality) ) {
        if ( any(res$species.quality[,ecoval.translate("R_macrophytes_error_type",dict)]==ecoval.translate("R_macrophytes_error_error",dict)) )
          cat(paste(outputLogger$errorPrefix, paste0(uiText$check.fail.species.data,":"),
            uiText$see, paste0(uiText$quality.checks,"\n")))
      }
      cat(paste0("END: ", Sys.time(), "\n"))
      return(res) # use to get site's codes and to interactively plot
    }
  )),
  readInputFilesAndEvaluate =
    function(dictionaryLanguage, inputDirData, inputDirImg, separator, inFileSites, inFileSpecies = NA, inFilePlausibleType = NA, samplingProtocol = macrophytesAPI$samplingProtocol$default, sampleSize = macrophytesAPI$defaultSampleSize, ...) {
    # Direct wrapper for `macrophytesAPI$.readInputFilesEvaluateAndWrite`, but
    # without writing, i.e. with `outputDir = NA`
      return(macrophytesAPI$readInputFilesEvaluateAndWrite(dictionaryLanguage, inputDirData, inputDirImg, separator, inFileSites, inFileSpecies = inFileSpecies, inFilePlausibleType = inFilePlausibleType, samplingProtocol = samplingProtocol, sampleSize = sampleSize, ...))
    }
  ,
  readInputFiles =
    function(dictionaryLanguage, inputDirData, inputDirImg, separator, inFileSites, inFileSpecies = NA, inFilePlausibleType = NA, samplingProtocol = macrophytesAPI$samplingProtocol$default, ...) {
    # Direct wrapper for `macrophytesAPI$.readInputFilesAndEvaluate`, but without
    # evaluation (much faster), i.e. with `sampleSize = 0`
      return(macrophytesAPI$readInputFilesAndEvaluate(dictionaryLanguage, inputDirData, inputDirImg, separator, inFileSites, inFileSpecies = inFileSpecies, inFilePlausibleType = inFilePlausibleType, samplingProtocol = samplingProtocol, sampleSize = 0, ...))
    }
  ,
  plotProbableTypes = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(...) {
      par(mar=c(0,0,0,0))
      msk.macrophytes.2017.plot.types.scheme(macrophytesAPI$.calcTypesOfObservation(...), 1)
    }
  )),
  plotProbableGrowthForms = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(...) {
      msk.macrophytes.2017.plot.types.grfo(macrophytesAPI$.calcTypesOfObservation(...), 1)
    }
  )),
  plotAssesmentOriginal = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(observationCode, res=NULL, ...) {
      # Call first `macrophytesAPI$readInputFilesEvaluateAndWrite()`, or
      # `macrophytesAPI$readInputFilesAndEvaluate()`, with `inFileSpecies` given
      if (!macrophytesAPI$hasAssesment(res))
        stop(uiText$assesment.prelim.first)
      # par(mar=c(0,0,0,0))
      # msk.macrophytes.2017.plot.hierarchy(res, macrophytesAPI$.getObservationCodeIdx(res, observationCode), final=FALSE, with.attrib=FALSE, cex.nodes=0.7)
      rivertype <- macrophytesAPI$getObservationRiverType(res, observationCode)
      macrophytesAPI$.plotAssesment(res, rivertype,
        macrophytesAPI$.calcValfunOfObservation(observationCode, res, rivertype),
        ...)
    }
  )),
  plotAssesmentPlausible = outputLogger$fwrap(.helpers$strtrimArgsWrap(
    function(observationCode, res=NULL, ...) {
      # Call first `macrophytesAPI$readInputFilesEvaluateAndWrite()`, or
      # `macrophytesAPI$readInputFilesAndEvaluate()`, with `inFileSpecies` and
      # `inFilePlausibleType` given
      if (!macrophytesAPI$hasAssesmentPlausible(res))
        stop(uiText$assesment.plaus.first)
      # par(mar=c(0,0,0,0))
      # msk.macrophytes.2017.plot.hierarchy(res, macrophytesAPI$.getObservationCodeIdx(res, observationCode), final=TRUE, with.attrib=FALSE, cex.nodes=0.7)
      rivertype <- macrophytesAPI$getObservationRiverTypeFinal(res, observationCode)
      macrophytesAPI$.plotAssesment(res, rivertype,
        macrophytesAPI$.calcValfunOfObservation(observationCode, res, rivertype),
        ...)
    }
  ))
)
