# server.R

if (!exists(".app.standalone") || is.na(as.logical(.app.standalone))) {
  .app.standalone <- FALSE
}

################################
#### Shiny app server       ####
################################
function(input, output, session) {



  ################################
  #config
  ################################
  stagingDir = file.path(getwd(), 'www')



  ################################
  #auxiliary functions
  ################################
  getFileURL.copyingIssueHandler <- function(cond) {
    message("Cannot copy output file to the staging folder: ",
      conditionMessage(cond))
    FALSE
  }
  getFileURL.createLinkingIssueHandler <- function(from, to) function(cond) {
    message("Cannot link output file in the staging folder: ",
      conditionMessage(cond))
    message("Copying output file instead.")
    handled <- tryCatch(
      {
        # no need to overwrite, because `unlink` should have been called first
        file.copy(from, to)
        TRUE
      },
      error=getFileURL.copyingIssueHandler,
      warning=getFileURL.copyingIssueHandler)
    print(paste("Linking issue handled?", handled))
    handled
  }
  getFileURL = function(filePath) {
    # Problem: linking to local server-side resources is disabled in all modern
    #          browsers due to security restrictions.
    # Workaround: (hard) link to the file in the safe "www/" folder (using
    #             file's basename) and return path of the hard link
    # TODO: consider a Windows-specific workaround - see: https://stackoverflow.com/a/14249649/207241
    linkPath <- basename(filePath)
    stagingPath <- file.path(stagingDir, linkPath)
    unlink(stagingPath)
    linkingIssueHandler <- getFileURL.createLinkingIssueHandler(filePath,
      stagingPath)
    isStaged <- tryCatch(
      {
        # fails e.g. when linking files between different Windows drives
        file.link(filePath, stagingPath)
        TRUE
      },
      error=linkingIssueHandler,
      warning=linkingIssueHandler)
    return(if (isStaged) linkPath else NA)
  }
  getFilesListURL = function(dirPath, filePattern="^.+\\.[^\\.]+") {

    # link all files with a given pattern
    links <- sapply(list.files(dirPath, filePattern, full.names = TRUE),
      getFileURL, USE.NAMES = FALSE)

    # prep HTML with index of all staged files (OPT: use modal popups instead)
    ulTag <- do.call(tags$ul,
      lapply(links, function(link) tags$li(tagPopoupLink(link))))
    docTag <- tags$html(tags$body(tags$h1(uiText$documentation), ulTag))

    # write HTML to file
    linkPath <- "doc.html"
    fileConn <- file(file.path(stagingDir, linkPath))
    # Note: overwrites previous content of the file
    writeLines(as.character(docTag), fileConn)
    close(fileConn)

    return(linkPath)
  }
  prepLogFile = function(logBaseName, resWrap) {
    # Note: using only `tempdir()` and not `tempfile()` to minimize amount of
    #       created files, because we don't remove them (yet?)
    logFile <- file.path(tempdir(), paste0(logBaseName,"Log.txt"))
    # Delete should not be necessary, as write overwrites previous contents
    # if (file.exists(logFile)) file.remove(logFile)
    write(outputLogger$getLogStr(resWrap), file=logFile)
    return(logFile)
  }



  tagBrokenLink = function(link.txt=uiText$broken.link) {
    tags$span(style="text-decoration: line-through;", link.txt)
  }
  tagPopoupLink = function(url, link.txt=NA) {
    if (is.na(link.txt)) link.txt = url
    if (is.na(url))
      tagBrokenLink(link.txt)
    else
      tags$a(target="_blank", href=url, link.txt)
  }
  tagFileLink = function(link.txt, file.name, dir.name) {
    url = getFileURL(file.path(dir.name, file.name))
    tagPopoupLink(url, link.txt)
  }
  tagFilesListLink = function(link.txt, file.name.pattern, dir.name) {
    url = getFilesListURL(dir.name, file.name.pattern)
    tagPopoupLink(url, link.txt)
  }

  tagAlertBox = function(type="info", ...)
    tags$div(class=paste0("alert alert-",type), ...)
  tagInfoBox = function(...) tagAlertBox(type="info", ...)
  tagWarnBox = function(...) tagAlertBox(type="warning", ...)
  tagDoneBox = function(...) tagAlertBox(type="success", ...)
  tagErrorBox = function(...) tagAlertBox(type="danger", ...)

  tagMsg = function(tagBox, msg, logFileURL=NA, more.txt=uiText$more.info)
    tagBox(
      tags$p(msg, if (is.na(logFileURL)) NULL
        else tags$a(target = "_blank", href = logFileURL, more.txt))
    )
  tagInfoMsg = function(info.msg, ...)
    tagMsg(tagInfoBox, info.msg, ...)
  tagWarnMsg = function(warn.msg, ...)
    tagMsg(tagWarnBox, warn.msg, ...)
  tagDoneMsg = function(done.msg=uiText$success, ...)
    tagMsg(tagDoneBox, done.msg, ...)
  tagErrorMsg = function(err.msg=uiText$error, ...)
    tagMsg(tagErrorBox, err.msg, ...)

  isInputValueMissing = function(value) {
    is.null(value) || .helpers$valueNotGiven(value)
  }
  hasInputValueChanged = function(oldValue, newValue) {
    if (!isInputValueMissing(oldValue))
      isInputValueMissing(newValue) || oldValue != newValue
    else # isInputValueMissing(oldValue)
      !isInputValueMissing(newValue)
  }
  validateInputGiven = function(inputName, alertOutputName, alertMsg,
    getInputValueFunc=getInputValue) {
    # check if sites data is given
    if (isInputValueMissing(getInputValueFunc(inputName))) {
      output[[alertOutputName]] <- renderUI(tagWarnMsg( alertMsg ))
      return(FALSE)
    }
    return(TRUE)
  }

# FIXME: re-design to use reactive elements in separate reactive functions, not
#        explicit updates calls
#   * temp. disabled validation of river types and attributes because
#     `update*Input` functions seemed no to affect input[[inputName]] value in
#     the same top-level reactive function call chain
#   * caching of inputs w/ getters is another, more general workaround (cf. 7dff443)
#   * cf. https://stackoverflow.com/a/40622303/207241
  getInputValue = function(inputName) {
    return(input[[inputName]])
  }
  updateNumericInputWrapper = function(inputName, value) {
    dispValue = input[[inputName]]
    updateNumericInput(session, inputName, value=value)
    return(list(previous=dispValue, current=value))
  }
  isInputValueValidChoice = function(value, choices) {
    !isInputValueMissing(value) && (value %in% choices)
  }
  updateSelectInputWrapper = function(inputName, choices=NULL, selected=NULL,
    auto.select.first = TRUE, ...)
  {
    dispValue = getInputValue(inputName)
    if (is.null(choices)) {
      choices = character(0)
      # ignore also selected value if no choices were given
      selected = ""
    } else if (!((is.character(selected) && (selected == "")) || isInputValueValidChoice(selected, choices))) {
      # if not a valid selection, auto-select first choice by default
      selected = if (isInputValueValidChoice(dispValue, choices)) dispValue
        else if (auto.select.first) head(choices, 1) else ""
    }
    updateSelectizeInput(session, inputName, choices=choices, selected=selected,
      ...)
    # return previous and selected input values
    return(list(previous=dispValue, current=selected))
  }

  getDirectoryInputValue = function(inputName)
    readDirectoryInput(session, inputName)
  chooseDirWrapper = function(inputName)
    choose.dir(default=getDirectoryInputValue(inputName),
      caption=uiText$select.directory)
  updateDirectoryInputWrapper = function(inputName) {
    updated <- FALSE
    path <- NA
    # condition prevents handler execution on initial app launch
    if (getInputValue(inputName) > 0) {
      path <- chooseDirWrapper(inputName)
      updated <- !is.na(path)
      if (updated)
        updateDirectoryInput(session, inputName, value=path)
    }
    return(list(updated=updated, path=path))
  }


  # temporarily swap running box and action output info boxes on each click
  actionRunningOff = function(actionPrefix) {
    shinyjs::hide(id=paste0(actionPrefix,"Running"))
    shinyjs::show(id=paste0(actionPrefix,"Output"))
  }
  actionRunningOn = function(actionPrefix) {
    shinyjs::show(id=paste0(actionPrefix,"Running"))
    shinyjs::hide(id=paste0(actionPrefix,"Output"))
  }


  updateOutLink = function(done.before, resOutputName, link.txt, linkFun, fn,
    ...)
  isolate({
    if (!isInputValueMissing(fn))
      output[[resOutputName]] <- renderUI(tags$p(
        tags$strong(linkFun(link.txt, fn, ...))
      ))
    else if (!done.before)
      output[[resOutputName]] <- renderUI(tags$p(link.txt))
  })
  updateOutFileLink = function(done.before, resOutputName, out.dir, fn,
    link.txt)
    updateOutLink(done.before, resOutputName, link.txt, tagFileLink, fn,
      out.dir)
  updateOutFilesListLink = function(done.before, resOutputName, out.dir,
    fn.pattern, link.txt)
    updateOutLink(done.before, resOutputName, link.txt, tagFilesListLink,
      fn.pattern, out.dir)





  ################################
  # lakes
  ################################
  ## State shared between actions
  # probably one should use here reactiveValues()
  lakes.doneOutput = list(
    "lakesReadFilesAndEvaluateOutput" = FALSE
  )
  lakes.attributes <- NULL
  lakesUpdateResults = function(res) {
    lakes.attributes <<- res
  }
  lakesGetResults = function() {
    lakes.attributes
  }
  lakeResultsAvail = function(res = lakesGetResults()) {
    #print(paste0("lakeResultsAvail: ", !is.null(res)))
    !is.null(res)
  }

  observeEvent(input$lakesInputDir, {
    updateDirectoryInputWrapper('lakesInputDir')
  })
  observeEvent(input$lakesOutputDir, {
    updateDirectoryInputWrapper('lakesOutputDir')
  })

  # results links
  lakesUpdateOutLinks = function(out.dir="", fn.attribute=NA, fn.assesment=NA,
    fn.hierarchies=NA, fn.evaluation.lines=NA)
  isolate({
    read.and.eval.done = lakeResultsAvail()
    updateOutFileLink(read.and.eval.done, "lakesAttributeOutput", out.dir,
      fn.attribute, uiText$attribute)
    updateOutFileLink(read.and.eval.done, "lakesAssesmentsOutput", out.dir,
      fn.assesment, uiText$assesments)
    updateOutFileLink(read.and.eval.done, "lakesHierarchiesOutput", out.dir,
      fn.hierarchies, uiText$target.hierarchies)
    updateOutFileLink(read.and.eval.done, "lakesEvaluationLinesOutput", out.dir,
      fn.evaluation.lines, uiText$evaluation.lines)
  })
  # initial text placeholders
  lakesUpdateOutLinks()

  # reset results incl. actions status and output links, but not the modifications buffer
  lakesActionOutputsReset = function(
    alertOutputNamesToReset = names(lakes.doneOutput),
    alertOutputNamesToTag = c(), tag = NULL,
    tag.otherwise = tagInfoMsg(HTML("&nbsp;")))
  {
    lakesUpdateResults(NULL)
    lakesUpdateOutLinks()
    # TODO: re-factor part below from here and macrophytesUpdateResults
    # unconditionally reset all outputs
    sapply(alertOutputNamesToReset, function(alertOutputName) {
      lakes.doneOutput[[alertOutputName]] <- FALSE
      alertOutputTag = if (!is.null(tag) && (alertOutputName %in% alertOutputNamesToTag))
        tag else tag.otherwise
      output[[alertOutputName]] <- renderUI(alertOutputTag)
    })
    NULL
  }
  # initial reset
  lakesActionOutputsReset()

  # tech: `onclick` event update comes only after UI update from observeEvent;
  #       using `mousedown` event instead
  shinyjs::onevent('mousedown', "lakesReadFilesAndEvaluateAction",
    actionRunningOn("lakesReadFilesAndEvaluate"))

  # write output file wrapper
  writeWrapper = function(outDir, apiFun, fileName, outFileExt, ...) {

    outFileName <- paste0(fileName, outFileExt)
    outFilePath <- file.path(outDir, outFileName)

    resWrap <- apiFun(outFilePath, ...)

    if (outputLogger$hasError(resWrap)) {
      # Note: not writing any logs to outDir, because we actually never
      #       checked that we can
      return(data.frame(success=FALSE, fn=I(prepLogFile(paste0("Lakes_",fileName), resWrap))))
    }

    return(data.frame(success=TRUE, fn=I(outFileName)))
  }

  # Read and evaluate on action button pressed
  observeEvent(input$lakesReadFilesAndEvaluateAction, isolate({

    # reset all previous outputs
    lakesActionOutputsReset()

    if (!validateInputGiven("lakesInputDir", "lakesReadFilesAndEvaluateOutput",
      uiText$input.dir.first, getInputValueFunc=getDirectoryInputValue)) {
      actionRunningOff("lakesReadFilesAndEvaluate")
      return(NULL)
    }

    if (!validateInputGiven("lakesOutputDir", "lakesReadFilesAndEvaluateOutput",
        uiText$output.dir.first, getInputValueFunc=getDirectoryInputValue)) {
      actionRunningOff("lakesReadFilesAndEvaluate")
      return(NULL)
    }

    inDir <- getDirectoryInputValue('lakesInputDir')
    outDir <- getDirectoryInputValue('lakesOutputDir')

    progress <- Progress$new(session, min=1, max=4)
    on.exit(progress$close())

    progress$set(message = paste(uiText$evaluating, uiText$please.wait, sep=".. "),
                detail = '')

    progress$set(value = 1)
    Sys.sleep(0.5)

    progress$set(value = 2)
    resWrap <- lakesAPI$readInputFiles(inDir,
      input$lakesColnameStartPoint, input$lakesColnameEndPoint, input$lakesColnameAttributeLevel, input$lakesColnameMeasureId,
      input$lakesColnameComments)
    logFileURL <- getFileURL(prepLogFile("Lakes_ReadInputFiles", resWrap))

    progress$set(value = 3)

    if (outputLogger$hasError(resWrap)) {
      render.tag <-
        tags$div(class = "alert alert-danger",
          tags$p(uiText$read.error,
            tags$a(target = "_blank", href = logFileURL, uiText$more.info)
          )
        )

    } else {

      res <- outputLogger$getValue(resWrap)

      # update global state
      # TODO: test if also: after UI update, otherwise links won't get updated
      lakesUpdateResults(res)
      lakes.doneOutput[["lakesReadFilesAndEvaluateOutput"]] <<- TRUE

      ## update UI

      # update segment choices
      lakesUpdateSegmentNameSelector(res)

      # # update sliders range
      # rangeVec <- lakesAPI$getSegmentsRange(res, input$lakesColnameStartPoint, input$lakesColnameEndPoint)
      # updateSliderInput(session, "hierarchienSlider", label = "Range", value = rangeVec,
      #                   min = rangeVec[1], max = rangeVec[2], step = NULL)
      # updateSliderInput(session, "bewertungsLinienSlider", label = "Range", value = rangeVec,
      #                   min = rangeVec[1], max = rangeVec[2], step = NULL)

      # write and update output links
      res.write.df <- NULL
      res.write.df <- rbind(res.write.df, writeWrapper(outDir,
        lakesAPI$writeAttributesFile,
        input$attributeFileName,
        lakesAPI$defaultFileExtensions$output.file.attributes,
        input$lakesOutputSeparator,
        input$lakesLanguage, res))
      res.write.df <- rbind(res.write.df, writeWrapper(outDir,
        lakesAPI$writeBewertungenFile,
        input$bewertungenFileName,
        lakesAPI$defaultFileExtensions$output.file.assesments,
        input$lakesOutputSeparator,
        input$lakesLanguage, res))
      res.write.df <- rbind(res.write.df, writeWrapper(outDir,
        lakesAPI$writeHierarchienFullFile,
        input$hierarchienFileName,
        lakesAPI$defaultFileExtensions$output.file.evaluation.hierarchies,
        input$hierarchienBreite, input$hierarchienHohe,
        NULL, #input$hierarchienSlider,
        input$lakesColnameStartPoint, input$lakesColnameEndPoint,
        input$lakesLanguage, res))
      res.write.df <- rbind(res.write.df, writeWrapper(outDir,
        lakesAPI$writeBewertungsLinienFile,
        input$bewertungsLinienFileName,
        lakesAPI$defaultFileExtensions$output.file.evaluation.lines,
        input$bewertungsLinienBreite, input$bewertungsLinienHohe,
        input$lakesColnameMeasureId,
        NULL, #input$bewertungsLinienSlider,
        input$lakesColnameStartPoint, input$lakesColnameEndPoint,
        input$lakesLanguage, res))

      lakesUpdateOutLinks(outDir,
        fn.attribute=if (res.write.df$success[1]) res.write.df$fn[1] else NA,
        fn.assesment=if (res.write.df$success[2]) res.write.df$fn[2] else NA,
        fn.hierarchies=if (res.write.df$success[3]) res.write.df$fn[3] else NA,
        fn.evaluation.lines=if (res.write.df$success[4]) res.write.df$fn[4] else NA)

      render.tag <- if (!all(res.write.df$success)) {
        ind.write.err <- min(which(!res.write.df$success))
        log.fn <- res.write.df$fn[ind.write.err]
        tags$div(class = "alert alert-danger",
          tags$p(uiText$write.error,
            tags$a(target = "_blank", href = getFileURL(log.fn), uiText$more.info)
          )
        )
      } else {
        tags$div(class = "alert alert-success",
          tags$p(
            uiText$read.success,
            tags$a(target = "_blank", href = logFileURL, uiText$more.info)
          )
        )
      }

      # Since there are observeEvents and chained updates there is no need to call
      # here either of:
      #     lakesUpdateModificationAttributes(res)
      #     lakesUpdatePlots(...)

    }

    progress$set(value = 4)

    output$lakesReadFilesAndEvaluateOutput <- renderUI(render.tag)
    actionRunningOff("lakesReadFilesAndEvaluate")
  }))

  lakesUpdateSegmentNameSelector = function(res) isolate({
    #print("lakesUpdateSegmentNameSelector")
    iv = updateSelectInputWrapper("lakesSegmentName",
      lakesAPI$getSegmentNames(res))
    # Since there is observeEvent on the observation code selector, call
    # explicitly attribute update only when update was triggered but observation
    # code hasn't changed (e.g. when re-computed evaluation)
    if (!hasInputValueChanged(iv$previous, iv$current))
      lakesUpdateModificationAttributes(res)
    NULL
  })

  # Re-plot on change of segment or option to switch on/off attributes (hierarchy levels)
  observeEvent({
      # works as OR over changes for all reactives inside of this `eventExpr`
      input$lakesSegmentName
      # despite the following isolate() over expressions, the explicit dep. on
      # input$lakesIfSingleHierarchyPlotWithAttrib is not needed, because this
      # (reactive) input implicitly registers renderPlot callback by having its
      # value accessed in the renderPlot's expressions; it would be needed if
      # input value would be accessed only outside of the renderPlot block
    }, isolate({
      #print("segment name changed - update plot")
      lakesUpdatePlots()
  }))

  lakesUpdatePlots = function() isolate({
    lakesPlotSingleHierarchy()
  })

  # Plot single hierachy
  lakesPlotSingleHierarchy = function() {
      if (!lakeResultsAvail()) {
        output$lakesSingleHierarchyPlot <- renderPlot({
          plotMsg(uiText$read.first)
        })
        # output$lakesSingleHierarchyOutput <- renderUI(
        #   tags$div(class = "alert alert-warning", tags$p(uiText$read.first)) )
        return()
      }
      # first, clean any previous status messages
      # output$lakesSingleHierarchyOutput <- renderUI(tags$div())
      # then render the plot, or just render error eport, if one occured
      output$lakesSingleHierarchyPlot <- renderPlot({
        resWrap <- lakesAPI$plotSingleHierarchy(
          input$lakesSegmentName,
          input$lakesLanguage,
          lakesGetResults(),
          with.attrib=getInputValue("lakesIfSingleHierarchyPlotWithAttrib")
        )
        # # DEBUG print call log
        # output[["lakesSingleHierarchyOutput"]] <- renderUI(
        #   tagInfoMsg(info.msg="DEBUG",
        #     logFileURL=getFileURL(prepLogFile("lakesPlotSingleHierarchy", resWrap)))
        # )
        if (outputLogger$hasError(resWrap)) {
          plotMsg(outputLogger$getError(resWrap))
          # output$lakesSingleHierarchyOutput <- renderUI(
          #   tags$div(class = "alert alert-danger",
          #     tags$p(uiText$plot.error,
          #       tags$a(target = "_blank", href = getFileURL(prepLogFile("Lakes_PlotSingleSegment", resWrap)), uiText$more.info)
          #     )
          #   )
          # )
        }
      })
  }

  # Update modification attributes on different segment selection
  observeEvent(input$lakesSegmentName, isolate({
    lakesUpdateModificationAttributes(lakesGetResults())
  }))

  #' Update all modfiable attribiutes inputs based on the API buffer and results
  #' object
  lakesUpdateModificationAttributes = function(res) isolate({
    lakesUpdateSingleHierarchyAttribute(res)
  })

  #' Update assesment attribiutes inputs (name and value) based on the API
  #' buffer and results object
  lakesUpdateSingleHierarchyAttribute = function(res) isolate({
    iv = updateSelectInputWrapper("lakesSingleHierarchyAttribute",
      lakesAPI$getHierarchyLevels(res))
    # Since there is observeEvent on the hierarchy attribute selector, call
    # explicitly attribute value update only when update was triggered but
    # attribute hasn't changed (e.g. when chosen a different observation)
    if (!hasInputValueChanged(iv$previous, iv$current))
      lakesUpdateSingleHierarchyAttributeValue(res)
    NULL
  })

  # Update hierarchy values to choose on change of hierarchy level
  observeEvent(input$lakesSingleHierarchyAttribute, isolate({
    lakesUpdateSingleHierarchyAttributeValue(lakesGetResults())
  }))

  #' Update assesment attribiute value input based on the API buffer and
  #' results object
  lakesUpdateSingleHierarchyAttributeValue = function(res) isolate({
    if (lakeResultsAvail(res) &&
        !isInputValueMissing(getInputValue("lakesSingleHierarchyAttribute")))
    {
      segment = getInputValue("lakesSegmentName")
      attrib = getInputValue("lakesSingleHierarchyAttribute")
      value = lakesAPI$getHierarchyLevelValueForSegment(res, segment, attrib)
      # # DEBUG
      # print("lakesUpdateSingleHierarchyAttributeValue")
      # print(segment)
      # print(attrib)
      # print(value)
      iv = updateSelectInputWrapper(
        "lakesSingleHierarchyAttributeValue",
        choices = lakesAPI$getHierarchyLevelValuesForHierarchyLevel(res, attrib),
        selected = value
      )
      # Since there is observeEvent on the hierachy lvl attribute value, call
      # explicitly plot update only when update was triggered but attribute
      # value hasn't changed (e.g. when re-computed evaluation)
      if (!hasInputValueChanged(iv$previous, iv$current))
        lakesUpdatePlots()
    }
    NULL
  })

  # Re-plot hierarchy on change of value of hierarchy attribute
  observeEvent(input$lakesSingleHierarchyAttributeValue, isolate({
    lakesSingleHierarchyModifyAttributeValue(lakesGetResults())
    lakesPlotSingleHierarchy()
  }))

  lakesSingleHierarchyModifyAttributeValue = function(res) {
    segment = getInputValue("lakesSegmentName")
    attrib = getInputValue("lakesSingleHierarchyAttribute")
    value = getInputValue("lakesSingleHierarchyAttributeValue")
    # # DEBUG
    # print("lakesSingleHierarchyModifyAttributeValue")
    # print(segment)
    # print(attrib)
    # print(value)
    if (lakeResultsAvail(res) &&
        !isInputValueMissing(segment) &&
        !isInputValueMissing(attrib) &&
        !isInputValueMissing(value))
    {
      lakesAPI$setHierarchyLevelValueForSegment(res, segment, attrib, value)
    }
  }

  # Reset interactive plot modifications
  observeEvent({ input$lakesClearModifications }, isolate({
      lakesResetModifications(lakesGetResults())
  }))

  lakesResetModifications = function(res) {
      # since observation code stays same, explicitly clear the mod. buffer,
      # _before_ updating the modification attributes based on computed results
      lakesAPI$modBuffer$clear()
      # reset inputs - buffer is empty so only data in cached results object
      # will be used
      lakesUpdateModificationAttributes(res)
  }

  ## END lakes




  ################################
  # macrophytes
  ################################
  ## State shared between actions
  ##
  # probably one should use here reactiveValues()
  macrophytes.doneOutput = list(
    "macrophytesCheckFilesOutput" = FALSE,
    "macrophytesComputeTypificationOutput" = FALSE,
    "macrophytesEvaluatePrelimOutput" = FALSE,
    "macrophytesEvaluatePlausOutput" = FALSE
  )
  macrophytes.checkAndEvalRes = NULL
  macrophytesUpdateResults = function(res) {
    macrophytes.checkAndEvalRes <<- res
  }

  ## Helpers
  ##
  # input files and directories
  macrophytes.fileNameInputs = c(
    "macrophytesSitesFileName",
    "macrophytesSpeciesFileName",
    "macrophytesPlausibilityFileName")
  macrophytes.fileNameChoices = c()
  macrophytesUpdateInputFileNames = function(dirPath=NA, inputs=macrophytes.fileNameInputs) {
    dirInput = getInputValue("macrophytesInputDataDir")
    if (dirInput > 0) {
      if (is.na(dirPath))
        dirPath = getDirectoryInputValue('macrophytesInputDataDir')
      file.names = list.files(dirPath, "\\.[^\\.]+$", full.names = FALSE)
      # update only if choices really changed
      if (length(macrophytes.fileNameChoices) != length(file.names) ||
          any(macrophytes.fileNameChoices != file.names))
      {
        macrophytes.fileNameChoices <<- file.names
        for (input in inputs) {
          updateSelectInputWrapper(input, file.names, auto.select.first=FALSE,
            server=TRUE)
          # load choices form server (server=TRUE), otherwise the dropdown will
          # close immediately after updateSelectizeInput call and
          # onevent("focusin",...) update (below) won't be usable
        }
      }
    }
  }
  # be careful: `for (fn in macrophytes.fileNameInputs)` won't work here
  shinyjs::onevent("focusin", "macrophytesSitesFileName",
      macrophytesUpdateInputFileNames())
  shinyjs::onevent("focusin", "macrophytesSpeciesFileName",
      macrophytesUpdateInputFileNames())
  shinyjs::onevent("focusin", "macrophytesPlausibilityFileName",
      macrophytesUpdateInputFileNames())

  observeEvent(input$macrophytesInputDataDir, {
    dir = updateDirectoryInputWrapper('macrophytesInputDataDir')
    if (dir$updated) {
      # on dir change: reset selections if files do not exist in the new dir
      # pass directly new path as the input's value might not be up to date yet
      macrophytesUpdateInputFileNames(dirPath=dir$path)
    }
  })
  observeEvent(input$macrophytesInputImgDir, {
    updateDirectoryInputWrapper('macrophytesInputImgDir')
  })
  observeEvent(input$macrophytesOutputDir, {
    updateDirectoryInputWrapper('macrophytesOutputDir')
  })

  # results links
  macrophytesUpdateOutLinks = function(out.dir="", fn.numeric=NA, fn.used.taxa=NA, fn.discarded.taxa=NA, fn.quality.checks=NA,
    fn.documentation=NA)
  isolate({
    read.done = !is.null(macrophytes.checkAndEvalRes)
    doc.done = read.done && getInputValue("macrophytesIfWriteDoc")
    eval.done = macrophytesAPI$hasTypification(macrophytes.checkAndEvalRes)
    updateOutFileLink(eval.done, "macrophytesNumResOutput", out.dir,
      fn.numeric, uiText$numeric)
    updateOutFileLink(eval.done, "macrophytesUsedTaxaOutput", out.dir,
      fn.used.taxa, uiText$used.taxa)
    updateOutFileLink(eval.done, "macrophytesDiscardedTaxaOutput", out.dir,
      fn.discarded.taxa, uiText$discarded.taxa)
    updateOutFileLink(read.done, "macrophytesQualityChecksOutput", out.dir,
      fn.quality.checks, uiText$quality.checks)
    fn.doc.pattern = if (is.na(fn.documentation)) NA else paste0("^.+",fn.documentation)
    updateOutFilesListLink(doc.done, "macrophytesDocumentationOutput",
      out.dir, fn.doc.pattern, uiText$documentation
    )
  })
  # initial text placeholders
  macrophytesUpdateOutLinks()

  # validate inputs pre check action
  macrophytesValidatePreCheck = function(alertOutputName) {
    res = list()

    # check if input dirs are chosen
    if (!validateInputGiven("macrophytesInputDataDir", alertOutputName,
      uiText$input.dir.first, getInputValueFunc=getDirectoryInputValue))
      return(NULL)
    # if (getInputValue("macrophytesIfWriteDoc") &&
    #   !validateInputGiven("macrophytesInputImgDir", alertOutputName,
    #     uiText$input.dir.first, getInputValueFunc=getDirectoryInputValue))
    #   return(NULL)
    # TODO: reset macrophytes.doneOutput on change of one of the input
    #       dirs or input files names, language, sampling protocol etc. ?
    res$inputDataDirPath <- getDirectoryInputValue("macrophytesInputDataDir")
    res$inputImgDirPath <- getDirectoryInputValue("macrophytesInputImgDir")

    # check if sites data is given
    if (!validateInputGiven("macrophytesSitesFileName", alertOutputName,
      paste(uiText$missing.filename, uiText$sites.data, sep=": "))) return(NULL)

    # check if the output dir is chosen
    if (!validateInputGiven("macrophytesOutputDir", alertOutputName,
      uiText$output.dir.first, getInputValueFunc=getDirectoryInputValue))
      return(NULL)
    res$outputDirPath <- getDirectoryInputValue("macrophytesOutputDir")

    return(res)
  }

  # validate inputs pre evaluation actions
  macrophytesValidatePreEvaluate = function(alertOutputName, sampling=TRUE, ...) {
    # OPT, require separate, quicker data check prior to actual evaluation
    # if (sampling)
    #   # check first if the input data check was successfully done
    #   if (!macrophytes.doneOutput[["macrophytesCheckFilesOutput"]]) {
    #     output[[alertOutputName]] <- renderUI(tagWarnBox(uiText$check.first))
    #     return(NULL)
    #   }

    return(macrophytesValidatePreCheck(alertOutputName))
  }

  macrophytesPrepareEvaluationArgs = function(dirs, sampling=TRUE,
    species=NULL, plaus=NULL) {
    if (is.null(species))
      species = !isInputValueMissing(getInputValue("macrophytesSpeciesFileName"))
    if (is.null(plaus))
      plaus = !isInputValueMissing(getInputValue("macrophytesPlausibilityFileName"))
    list(
      getInputValue("macrophytesLanguage"),
      dirs$inputDataDirPath,
      dirs$inputImgDirPath,
      getInputValue("macrophytesInputSeparator"),
      .helpers$appendFnExtIfMissing( # inFileSites
        getInputValue("macrophytesSitesFileName"),
        macrophytesAPI$defaultFileExtensions$input.file.site),
      samplingProtocol = getInputValue("macrophytesSamplingProtocol"),
      outputDir = dirs$outputDirPath,
      outputSeparator = getInputValue("macrophytesOutputSeparator"),
      outFileQualityChecks = .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesQualityChecksFileName"),
        macrophytesAPI$defaultFileExtensions$output.file.quality.checks),
      outFileDoc = if (getInputValue("macrophytesIfWriteDoc")) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesDocFileName"),
        macrophytesAPI$defaultFileExtensions$output.file.doc) else NA,
      # sampling-dependent
      sampleSize = if (sampling) as.numeric(getInputValue("macrophytesNSamples")) else 0,
      outFileNumeric = if (sampling) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesNumResFileName"),
        macrophytesAPI$defaultFileExtensions$output.file.numeric) else NA,
      # species-dependent
      inFileSpecies = if (species) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesSpeciesFileName"),
        macrophytesAPI$defaultFileExtensions$input.file.species) else NA,
      outFileTaxaUsed = if (species) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesUsedTaxaFileName"),
        macrophytesAPI$defaultFileExtensions$output.file.taxa.used) else NA,
      outFileTaxaRemoved = if (species) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesDiscardedTaxaFileName"),
        macrophytesAPI$defaultFileExtensions$output.file.taxa.removed) else NA,
      # plaus-dependent
      inFilePlausibleType = if (plaus) .helpers$appendFnExtIfMissing(
        getInputValue("macrophytesPlausibilityFileName"),
        macrophytesAPI$defaultFileExtensions$input.file.typeplaus) else NA
    )
  }

  macrophytesEvaluateWrapper = function(alertOutputName, ...) {
    # validate state pre-evaluate
    valid <- macrophytesValidatePreEvaluate(alertOutputName, ...)
    if (is.null(valid)) return(FALSE)

    # evaluate
    args = macrophytesPrepareEvaluationArgs(valid, ...)
    #n.in = length(macrophytesAPI$getObservationCodes(macrophytes.checkAndEvalRes))
    #n.logsamp = max(0,log10(args$sampleSize)-1)
    #t.wait=ceiling(5+0.3*n.in*n.logsamp) # w/ n.in=50, for n.samp = 0: 5s, 100: 20s, 1000: 35s, 10000: 50s
    progress <- Progress$new(session, min=1, max=3)
    on.exit(progress$close())
    progress$set(message = paste(uiText$evaluating, uiText$please.wait, sep=".. "),
                 detail = '')
    progress$set(value = 1)
    Sys.sleep(0.5)
    progress$set(value = 2)
    resWrap <- do.call(macrophytesAPI$readInputFilesEvaluateAndWrite, args)
    logFileURL <- getFileURL(prepLogFile(alertOutputName, resWrap))
    progress$set(value = 3)

    # check for errors
    if (outputLogger$hasError(resWrap)) {
      output[[alertOutputName]] <- renderUI(tagErrorMsg(logFileURL=logFileURL))
      return(FALSE)
    }

    res <- outputLogger$getValue(resWrap)

    # explicitly reset intereactive changes buffer
    macrophytesAPI$modBuffer.new()

    # update UI
    macrophytesUpdateOutLinks(valid$outputDirPath, fn.numeric = args$outFileNumeric,
      fn.used.taxa = args$outFileTaxaUsed,
      fn.discarded.taxa = args$outFileTaxaRemoved,
      fn.quality.checks = args$outFileQualityChecks,
      fn.documentation = args$outFileDoc)
    macrophytesUpdateObservationCodeSelector(res)
    # Since there are observeEvents and chained updates there is no need to call
    # here either of:
    #     macrophytesUpdateModificationAttributes(res)
    #     macrophytesUpdatePlotsFromInputs(...)

    # update global state (after UI update, otherwise links won't get updated)
    macrophytesUpdateResults(res)
    macrophytes.doneOutput[[alertOutputName]] <<- TRUE

    # write success or non-fatal error msg, and link the log
    nonFatalError = outputLogger$hasLogError(resWrap)
    output[[alertOutputName]] <- if (nonFatalError)
      renderUI(tagErrorMsg(logFileURL=logFileURL))
      else renderUI(tagDoneMsg(logFileURL=logFileURL))

    return(TRUE)
  }

  # reset results incl. actions status and output links, but not the modifications buffer
  macrophytesActionOutputsReset = function(
    alertOutputNamesToReset = names(macrophytes.doneOutput),
    alertOutputNamesToTag = c(), tag = NULL,
    tag.otherwise = tagInfoMsg(HTML("&nbsp;")))
  {
    macrophytesUpdateResults(NULL)
    macrophytesUpdateOutLinks() # unconditionally reset all outputs
    sapply(alertOutputNamesToReset, function(alertOutputName) {
      macrophytes.doneOutput[[alertOutputName]] <- FALSE
      alertOutputTag = if (!is.null(tag) && (alertOutputName %in% alertOutputNamesToTag))
        tag else tag.otherwise
      output[[alertOutputName]] <- renderUI(alertOutputTag)
    })
    NULL
  }
  # initial reset
  macrophytesActionOutputsReset()


  ####BEGIN: evaluations
  # tech: `onclick` event update comes only after UI update from observeEvent;
  #       using `mousedown` event instead
  shinyjs::onevent('mousedown', "macrophytesCheckFilesAction",
    actionRunningOn("macrophytesCheckFiles"))
  shinyjs::onevent('mousedown', "macrophytesComputeTypificationAction",
    actionRunningOn("macrophytesComputeTypification"))
  shinyjs::onevent('mousedown', "macrophytesEvaluatePrelimAction",
    actionRunningOn("macrophytesEvaluatePrelim"))
  shinyjs::onevent('mousedown',"macrophytesEvaluatePlausAction",
    actionRunningOn("macrophytesEvaluatePlaus"))
  observeEvent(input$macrophytesCheckFilesAction, isolate({
    # reset all previous outputs
    macrophytesActionOutputsReset()
    alertOutputName = "macrophytesCheckFilesOutput"
    # eval w/o sampling
    eval_ok = macrophytesEvaluateWrapper(alertOutputName, sampling=FALSE)
    if (eval_ok) macrophytesUpdatePlotsFromInputs()
    actionRunningOff("macrophytesCheckFiles")
  }))
  observeEvent(input$macrophytesComputeTypificationAction, isolate({
    # reset all previous outputs
    macrophytesActionOutputsReset()
    alertOutputName = "macrophytesComputeTypificationOutput"
    eval_ok = macrophytesEvaluateWrapper(alertOutputName, species=FALSE, plaus=FALSE)
    if (eval_ok) macrophytesUpdatePlotsFromInputs()
    actionRunningOff("macrophytesComputeTypification")
  }))
  observeEvent(input$macrophytesEvaluatePrelimAction, isolate({
    # reset all previous outputs
    macrophytesActionOutputsReset()
    alertOutputName = "macrophytesEvaluatePrelimOutput"
    # check if species data is given
    if (validateInputGiven("macrophytesSpeciesFileName", alertOutputName,
      paste(uiText$missing.filename, uiText$species.data, sep=": "))) {
      eval_ok = macrophytesEvaluateWrapper(alertOutputName, species=TRUE, plaus=FALSE)
      if (eval_ok) macrophytesUpdatePlotsFromInputs()
    }
    actionRunningOff("macrophytesEvaluatePrelim")
  }))
  observeEvent(input$macrophytesEvaluatePlausAction, isolate({
    # reset all previous outputs
    macrophytesActionOutputsReset()
    alertOutputName = "macrophytesEvaluatePlausOutput"
    # check if species and plausibility data are given
    if (
      validateInputGiven("macrophytesSpeciesFileName", alertOutputName,
        paste(uiText$missing.filename, uiText$species.data, sep=": ")) &&
      validateInputGiven("macrophytesPlausibilityFileName", alertOutputName,
        paste(uiText$missing.filename, uiText$plausibilised.river.types, sep=": "))
    ) {
      eval_ok = macrophytesEvaluateWrapper(alertOutputName, species=TRUE, plaus=TRUE)
      if (eval_ok) macrophytesUpdatePlotsFromInputs()
    }
    actionRunningOff("macrophytesEvaluatePlaus")
  }))
  ###END onclick: evaluations


  #### BEGIN: interactive plots
  ## interactive modifications: update inputs
  ##
  macrophytesUpdateObservationCodeSelector = function(res) isolate({
    iv = updateSelectInputWrapper("macrophytesObservationCode",
      macrophytesAPI$getObservationCodes(res))
    # Since there is observeEvent on the observation code selector, call
    # explicitly attribute update only when update was triggered but observation
    # code hasn't changed (e.g. when re-computed typification)
    if (!hasInputValueChanged(iv$previous, iv$current))
      macrophytesUpdateModificationAttributes(res)
    NULL
  })

  #' Update all modfiable attribiutes inputs based on the API buffer and results
  #' object
  macrophytesUpdateModificationAttributes = function(res) isolate({
    macrophytesUpdateTypificationAttribute(res)
    macrophytesUpdateAssesmentAttribute(res)
  })
  #' Update typfication attribiutes inputs (name and value) based on the API
  #' buffer and results object
  macrophytesUpdateTypificationAttribute = function(res) isolate({
    iv = updateSelectInputWrapper("macrophytesTypificationAttribute",
      macrophytesAPI$getAttributeTypes(res))
    # Since there is observeEvent on the attribute selector, call explicitly
    # attribute value update only when update was triggered but attribute
    # hasn't changed (e.g. when chosen a different observation)
    # or
    # when the SelectInput hasn't yest change input (because of isolate?)
    if (!hasInputValueChanged(iv$previous, iv$current) ||
      hasInputValueChanged(getInputValue("macrophytesTypificationAttribute"), iv$current))
    {
      macrophytesUpdateTypificationAttributeValue(res, typification.attribute=iv$current)
    }
    NULL
  })
  #' Update typfication attribiute value input based on the API buffer and
  #' results object
  macrophytesUpdateTypificationAttributeValue = function(res,
    typification.attribute=NULL, update.plot=NULL)
  isolate({
    if (is.null(typification.attribute))
      typification.attribute = getInputValue("macrophytesTypificationAttribute")
    inputName = "macrophytesTypificationAttributeValue"
    inputValue = getInputValue(inputName)
    if (!is.null(res) && !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(typification.attribute))
    {
      newValue = macrophytesAPI$getObservationAttributeTypeValue(res,
        getInputValue("macrophytesObservationCode"),
        typification.attribute)
      # OPT: change `min`/`max` and `step` kwargs here dep. on the
      #      `typification.attribute` value
      iv = updateNumericInputWrapper(inputName, newValue)
      # print(iv)
      # There is observeEvent on the typification attribute value
      if.update.plot = ifUpdatePlotDespiteObserveEvent(iv$previous, iv$current,
        update.now=update.plot)
      iv[["update.plot.pending"]] = if.update.plot$pending
      if (if.update.plot$now)
        macrophytesUpdateTypificationPlots(getInputValue("macrophytesTypificationPlotChoice"))
    } else {
      iv = list(previous=inputValue, current=inputValue, update.plot.pending=FALSE)
    }
    iv
  })
  #' Update assesment attribiutes inputs (name and value) based on the API
  #' buffer and results object
  macrophytesUpdateAssesmentAttribute = function(res) isolate({
    # print("macrophytesUpdateAssesmentAttribute")
    iv = updateSelectInputWrapper("macrophytesAssesmentAttribute",
      macrophytesAPI$getAttributeSpecies(res))
    # print(iv)
    # Since there is observeEvent on the attribute selector, call explicitly
    # attribute value update only when update was triggered but attribute
    # hasn't changed (e.g. when chosen a different observation)
    # or
    # when the SelectInput hasn't yet changed input (because of isolate?)
    iv.attr = list(update.plot.pending = NULL)
    if (!hasInputValueChanged(iv$previous, iv$current) ||
      hasInputValueChanged(getInputValue("macrophytesAssesmentAttribute"), iv$current))
    {
      # do not update plot now to avoid override of buffer with displayed input values
      # when clearing modifications - do it only in chained river type update
      iv.attr = macrophytesUpdateAssesmentAttributeValue(res,
        assesment.attribute=iv$current, update.plot=FALSE)
    }
    # chain update river type
    macrophytesUpdateAssesmentRiverType(res,
      update.plot=iv.attr[["update.plot.pending"]])
    NULL
  })
  #' Update assesment attribiute value input based on the API buffer and
  #' results object
  macrophytesUpdateAssesmentAttributeValue = function(res, assesment.attribute=NULL,
    update.plot=NULL)
  isolate({
    # print("macrophytesUpdateAssesmentAttributeValue")
    if (is.null(assesment.attribute))
      assesment.attribute = getInputValue("macrophytesAssesmentAttribute")
    inputName = "macrophytesAssesmentAttributeValue"
    inputValue = getInputValue(inputName)
    if (!is.null(res) && !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(assesment.attribute))
    {
      newValue = macrophytesAPI$getObservationAttributeSpeciesValue(res,
        getInputValue("macrophytesObservationCode"),
        assesment.attribute)
      # OPT: change `min`/`max` and `step` kwargs here dep. on the
      #      `assesment.attribute` value
      iv = updateNumericInputWrapper(inputName, newValue)
      # print(iv)
      # There is observeEvent on the assesment attribute value
      if.update.plot = ifUpdatePlotDespiteObserveEvent(iv$previous, iv$current,
        update.now=update.plot)
      iv[["update.plot.pending"]] = if.update.plot$pending
      if (if.update.plot$now)
        macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))

    } else {
      iv = list(previous=inputValue, current=inputValue, update.plot.pending=FALSE)
    }
    iv
  })
  #' Update river type name input based on the the API buffer and results object
  macrophytesUpdateAssesmentRiverType = function(res, update.plot=NULL) isolate({
    # print("macrophytesUpdateAssesmentRiverType")
    inputName = "macrophytesAssesmentRiverType"
    inputValue = getInputValue(inputName)
    if (!is.null(res) && !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(getInputValue("macrophytesAssesmentPlotChoice")))
    {
      observationCode = getInputValue("macrophytesObservationCode")
      assChoice = getInputValue("macrophytesAssesmentPlotChoice")
      selected = NULL
      if (macrophytesIsAssesmentPlotPrelim(assChoice))
        selected = macrophytesAPI$getObservationRiverType(res, observationCode)
      else if (macrophytesIsAssesmentPlotFinal(assChoice))
        selected = macrophytesAPI$getObservationRiverTypeFinal(res, observationCode)
      iv = updateSelectInputWrapper(inputName, macrophytesAPI$getRiverTypes(res),
        selected=selected)
      # print(iv)
      # There is observeEvent on the assesment river type value
      if.update.plot = ifUpdatePlotDespiteObserveEvent(iv$previous, iv$current,
        update.now=update.plot)
      iv[["update.plot.pending"]] = if.update.plot$pending
      if (if.update.plot$now)
        macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))
    } else {
      iv = list(previous=inputValue, current=inputValue, update.plot.pending=FALSE)
    }
    iv
  })
  ifUpdatePlotDespiteObserveEvent = function(input.previous, input.current, update.now=NULL) {
    # Use if there is observeEvent on the input change that wuold normally trigger plot
    # update, but you want to force update in case the value hasn't changed
    update.pending = FALSE
    if(!hasInputValueChanged(input.previous, input.current)) {
      if (is.null(update.now)) update.now = TRUE
      else if (!update.now) update.pending = TRUE
    } else if (is.null(update.now)) update.now = FALSE
    list(now=update.now, pending=update.pending)
  }

  ## interactive modifications: apply
  ##
  macrophytesTypificationModifyAttributeValue = function(res) {
    if (!is.null(res) &&
        !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(getInputValue("macrophytesTypificationAttribute")) &&
        !isInputValueMissing(getInputValue("macrophytesTypificationAttributeValue")))
    {
      macrophytesAPI$setObservationAttributeTypeValue(
        res, getInputValue("macrophytesObservationCode"),
        getInputValue("macrophytesTypificationAttribute"),
        getInputValue("macrophytesTypificationAttributeValue")
      )
    }
  }
  macrophytesAssesmentModifyAttributeValue = function(res) {
    # print("macrophytesAssesmentModifyAttributeValue")
    # for (inputName in c("macrophytesObservationCode", "macrophytesAssesmentAttribute", "macrophytesAssesmentAttributeValue")) {
    #   iv = getInputValue(inputName)
    #   print(paste("    Is", inputName, "missing?", isInputValueMissing(iv), "value =", iv))
    # }
    if (!is.null(res) &&
        !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(getInputValue("macrophytesAssesmentAttribute")) &&
        !isInputValueMissing(getInputValue("macrophytesAssesmentAttributeValue")))
    {
      macrophytesAPI$setObservationAttributeSpeciesValue(
        res, getInputValue("macrophytesObservationCode"),
        getInputValue("macrophytesAssesmentAttribute"),
        getInputValue("macrophytesAssesmentAttributeValue")
      )
    }
    # chained update on river type here (because plot wrapper calls
    # "...ModifyAttributeValue" only)
    return(macrophytesAssesmentModifyRiverType(res))
  }
  macrophytesAssesmentModifyRiverType = function(res) {
    # print("macrophytesAssesmentModifyRiverType")
    if (!is.null(res) &&
        !isInputValueMissing(getInputValue("macrophytesObservationCode")) &&
        !isInputValueMissing(getInputValue("macrophytesAssesmentRiverType")))
    {
      observationCode = getInputValue("macrophytesObservationCode")
      assChoice = getInputValue("macrophytesAssesmentPlotChoice")
      if (macrophytesIsAssesmentPlotPrelim(assChoice))
          macrophytesAPI$setObservationRiverType(
            res, getInputValue("macrophytesObservationCode"),
            getInputValue("macrophytesAssesmentRiverType"))
      else if (macrophytesIsAssesmentPlotFinal(assChoice))
          macrophytesAPI$setObservationRiverTypeFinal(
            res, getInputValue("macrophytesObservationCode"),
            getInputValue("macrophytesAssesmentRiverType"))
    }
  }


  plotMsg = function(msg) {
      plot.new()
      text(.5,.5, msg)
  }
  macrophytesPreparePlotArgs = function(modifyAttrFun, ...) {
    # first, modify attributes for the interactive plot
    modifyAttrFun(macrophytes.checkAndEvalRes)
    list(
      getInputValue("macrophytesObservationCode"),
      macrophytes.checkAndEvalRes,
      sampleSize = as.numeric(getInputValue("macrophytesNSamples")),
      dictionaryLanguage = getInputValue("macrophytesLanguage"),
      ...)
  }
  macrophytesPlot = function(uiPrefix, apiFunName, ...) {
    alertOutputName = paste0(uiPrefix,"MessageOutput")
    plotOutputName = paste0(uiPrefix,"Plot")
    modifyAttrFun = base::get(paste0(uiPrefix,"ModifyAttributeValue"))
    plotFun = macrophytesAPI[[apiFunName]]
    # clean any previous status messages
    output[[alertOutputName]] <- renderUI(tags$div())
    # render the plot (prevent accidental dependencies)
    args = macrophytesPreparePlotArgs(modifyAttrFun, ...)
    # DEBUG
    # print("macrophytesPlot")
    # print(plotOutputName)
    # print(apiFunName)
    # print(args)
    # cat("\n\n")
    # if (!isInputValueMissing(getInputValue("macrophytesTypificationAttribute")))
    #   print(macrophytesAPI$getObservationAttributeTypeValue(
    #       macrophytes.checkAndEvalRes, getInputValue("macrophytesObservationCode"),
    #       getInputValue("macrophytesTypificationAttribute")))
    # if (!isInputValueMissing(getInputValue("macrophytesAssesmentAttribute")))
    #   print(macrophytesAPI$getObservationAttributeSpeciesValue(
    #       macrophytes.checkAndEvalRes, getInputValue("macrophytesObservationCode"),
    #       getInputValue("macrophytesAssesmentAttribute")))
    # cat("\n\n")
    output[[plotOutputName]] <- renderPlot({
      resWrap <- do.call(plotFun, args)
      # DEBUG print call log
      # output[[alertOutputName]] <- renderUI(
      #   tagInfoMsg(info.msg="DEBUG",
      #     logFileURL=getFileURL(prepLogFile(alertOutputName, resWrap)))
      # )
      if (outputLogger$hasError(resWrap)) {
        plotMsg(outputLogger$getError(resWrap))
        # OPT, put error box instead of a dummy plot
        # output[[alertOutputName]] <- renderUI(
        #   tagErrorMsg(err.msg=uiText$plot.error,
        #     logFileURL=getFileURL(prepLogFile(alertOutputName, resWrap)))
        # )
      }
    })
  }


  macrophytesIsTypificationPlotTypes = function(typChoice)
    typChoice==uiText$types
  macrophytesIsTypificationPlotGrowthForms = function(typChoice)
    typChoice==uiText$growth.habit.group
  macrophytesUpdateTypificationPlots = function(typChoice) isolate({
    if (macrophytesIsTypificationPlotTypes(typChoice))
      macrophytesPlot("macrophytesTypification", "plotProbableTypes")
    else if (macrophytesIsTypificationPlotGrowthForms(typChoice))
      macrophytesPlot("macrophytesTypification", "plotProbableGrowthForms")
  })
  macrophytesIsAssesmentPlotPrelim = function(assChoice) assChoice==uiText$orig
  macrophytesIsAssesmentPlotFinal = function(assChoice) assChoice==uiText$plaus
  macrophytesUpdateAssesmentPlots = function(assChoice) isolate({
    # print("macrophytesUpdateAssesmentPlots")
    if (macrophytesIsAssesmentPlotPrelim(assChoice))
      macrophytesPlot("macrophytesAssesment", "plotAssesmentOriginal",
        with.attrib=getInputValue("macrophytesIfAssesmentPlotWithAttrib"))
    else if (macrophytesIsAssesmentPlotFinal(assChoice))
      macrophytesPlot("macrophytesAssesment", "plotAssesmentPlausible",
        with.attrib=getInputValue("macrophytesIfAssesmentPlotWithAttrib"))
  })
  macrophytesUpdatePlots = function(typChoice = uiText$types,
    assChoice = uiText$orig)
  isolate({
    # print("macrophytesUpdatePlots")
    macrophytesUpdateTypificationPlots(typChoice)
    macrophytesUpdateAssesmentPlots(assChoice)
    NULL
  })
  # generate dummies at server start
  macrophytesUpdatePlots()
  macrophytesUpdatePlotsFromInputs = function() {
    return(macrophytesUpdatePlots(getInputValue("macrophytesTypificationPlotChoice"),
        getInputValue("macrophytesAssesmentPlotChoice")))
  }


  ## auto-update plots or attributes values whenever choices in UI change
  ##
  observeEvent(input$macrophytesTypificationPlotChoice, isolate({
    # print("observeEvent: macrophytesTypificationPlotChoice")
    macrophytesUpdateTypificationPlots(getInputValue("macrophytesTypificationPlotChoice"))
  }))
  observeEvent(input$macrophytesAssesmentPlotChoice, isolate({
      # print("observeEvent: macrophytesAssesmentPlotChoice")
      # important: update possibly different river type input _before_ re-plotting
      macrophytesUpdateAssesmentRiverType(macrophytes.checkAndEvalRes)
      macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))
  }))
  observeEvent(input$macrophytesObservationCode, isolate({
    # print("observeEvent: macrophytesObservationCode")
    macrophytesUpdateModificationAttributes(macrophytes.checkAndEvalRes)
  }))
  observeEvent(input$macrophytesTypificationAttribute, isolate({
    # print("observeEvent: macrophytesTypificationAttribute")
    macrophytesUpdateTypificationAttributeValue(macrophytes.checkAndEvalRes)
  }))
  observeEvent(input$macrophytesTypificationAttributeValue, isolate({
    # print("observeEvent: macrophytesTypificationAttributeValue")
    macrophytesUpdateTypificationPlots(getInputValue("macrophytesTypificationPlotChoice"))
  }))
  observeEvent(input$macrophytesAssesmentAttribute, isolate({
    # print("observeEvent: macrophytesAssesmentAttribute")
    macrophytesUpdateAssesmentAttributeValue(macrophytes.checkAndEvalRes)
  }))
  observeEvent(input$macrophytesAssesmentAttributeValue, isolate({
    # print("observeEvent: macrophytesAssesmentAttributeValue")
    macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))
  }))
  observeEvent(input$macrophytesAssesmentRiverType, isolate({
    # print("observeEvent: macrophytesAssesmentRiverType")
    macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))
  }))
  observeEvent(input$macrophytesIfAssesmentPlotWithAttrib, isolate({
    # print("observeEvent: macrophytesIfAssesmentPlotWithAttrib")
    macrophytesUpdateAssesmentPlots(getInputValue("macrophytesAssesmentPlotChoice"))
  }))

  # reset interactive plot modifications
  macrophytesResetModifications = function(res) {
      # print("macrophytesResetModifications")
      # since observation code stays same, explicitly clear the mod. buffer,
      # _before_ updating the modification attributes based on computed results
      macrophytesAPI$modBuffer.new()
      # reset inputs - buffer is empty so only data in cached results object
      # will be used
      macrophytesUpdateModificationAttributes(res)
  }
  observeEvent({ input$macrophytesTypificationClearModifications }, isolate({
      # print("observeEvent: macrophytesTypificationClearModifications")
      macrophytesResetModifications(macrophytes.checkAndEvalRes)
  }))
  observeEvent({ input$macrophytesAssesmentClearModifications }, isolate({
      # print("observeEvent: macrophytesAssesmentClearModifications")
      macrophytesResetModifications(macrophytes.checkAndEvalRes)
  }))

  # Reset outputs, modifications and plots directly on language change
  observeEvent(input$macrophytesLanguage, isolate({
      # print("observeEvent: macrophytesLanguage")
      macrophytesActionOutputsReset()
      stopifnot(is.null(macrophytes.checkAndEvalRes))
      macrophytesResetModifications(macrophytes.checkAndEvalRes)
      macrophytesUpdatePlots()
  }))
  ####END: interactive plots


  if (!interactive()) {
    # Close the app when the session completes
    session$onSessionEnded(function() {
      stopApp()
      q("no")
    })
  }
}
