installed_pkgs = data.frame(installed.packages())

# Ensure that a package is installed
ensure <- function(pkg, pkg_name, repo = config$pkgs$cran, pkgs = pkg, check_connection = TRUE) {

  setProgressBar(pb,
                    value = grep(paste0("\\b", pkg_name, "\\b"), names(pkgs)) / (length(pkgs) + 1),
                    label = sprintf("Loading - %s...", pkg_name))

  # Get the CRAN version & requirements
  installed_version <- if (pkg_name %in% installed_pkgs$Package)
    installed_pkgs$Version[installed_pkgs$Package == pkg_name][1]
  else NA
  breakpoint <- attr(regexpr("[<>=]+", pkg), "match.length")
  inequality <- substr(pkg, 1, breakpoint)
  required_version <- substr(pkg, breakpoint + 1, nchar(pkg))
  message(pkg_name, " required version: ", required_version)

  # Check if the installed version meets the specs
  specs_met <- if_version_specs_met(installed_version, inequality, required_version)
  message(pkg_name, " already available: ", specs_met, paste0(
    " (required: ", inequality, required_version,
    "; installed: ", installed_version, ")"
  ))
  # Check if package is installed and the specs are met
  if (!specs_met) {
    if (check_connection) ensure_site_connection(repo)
    cran_version <- pkgVersionCRAN(pkg_name, repo = repo)
    specs_met <- if_version_specs_met(cran_version, inequality, required_version)
    if (specs_met) {
      install.packages(pkg_name, repos = repo, dependencies = TRUE)
    } else {
      if (!require("devtools"))
        stop(pkg_name, " version ", pkg, " cannot be installed ",
             "without the \"devtools\" package installed first.")
      devtools::install_version(pkg_name, version = required_version, repos = repo, dependencies = TRUE)
    }
    message(pkg_name,  " installed")
  }
  library(pkg_name, character.only = TRUE)
  message(pkg_name, " loaded\n")
  Sys.sleep(0.2)
}

if_version_specs_met <- function(checked_version, inequality, required_version) {
  specs_met <- if (is.na(checked_version)) {
    FALSE
  } else {
    eval(parse(
      text = paste0(
        "numeric_version('", checked_version, "')",
        inequality,
        "numeric_version('", required_version, "')"
      )
    ))
  }
  return (specs_met)
}

# Ensure that remotes are installed
ensure_remotes <- function(remote) {
  setProgressBar(pb,
                    value = grep(paste0("\\b", remote, "\\b"), remotes) / (length(remotes) + 1),
                    label = sprintf("Loading - %s...", remote))
  pkg <- basename(remote)
  if (!(pkg %in% row.names(installed.packages()))) {
    ensure_site_connection("https://github.com")
    devtools::install_github(remote, dependencies = TRUE)
  }
  library(pkg, character.only = TRUE)
  message(paste0(pkg_name, " installed\n"))
  Sys.sleep(0.2)
}

# Ensures local packages are installed
ensure_local <- function(pkg, pkg_name, lib_path = file.path(appwd, config$locals$local)) {
  setProgressBar(pb,
                    value = grep(paste0("\\b", pkg_name, "\\b"), names(locals)) / (length(locals) + 1),
                    label = sprintf("Loading - %s...", pkg_name))

  # Get the requirements
  installed_version <- installed_pkgs$Version[installed_pkgs$Package == pkg_name]
  breakpoint <- attr(regexpr("[<>=]+", pkg), "match.length")
  inequality <- substr(pkg, 1, breakpoint)
  required_version <- substr(pkg, breakpoint + 1, nchar(pkg))

  # Check if the installed version meets the specs
  if (length(installed_version) == 0) {
    specs_not_met <- TRUE
  } else {
    specs_not_met <- !eval(parse(text =
                                   paste0("numeric_version('", installed_version, "')",
                                          inequality,
                                          "numeric_version('", required_version, "')")))
  }
  if (!pkg_name %in% installed_pkgs$Package | specs_not_met) {
    install.packages(list.files(path = lib_path, pattern = pkg_name, full.names = TRUE), repos = NULL, type = "source")
    message(pkg_name, " installed")
  }
  library(pkg_name, character.only = TRUE)
  message(pkg_name, " loaded\n")
  Sys.sleep(0.2)
}

# Internet connection test
ensure_site_connection <- function(site_url) {
  if (!test_http(site_url)) {
    err_msg = sprintf("Can't connect to %s...", site_url)
    stop(err_msg)
  }
}
test_http <- function(site_url) {
  fail <- try(httr::http_error(site_url))
  return(!(inherits(fail, "try-error") || fail))
}

# Check CRAN if package version is current
pkgVersionCRAN = function(pkg_name, repo = "http://cran.r-project.org/") {

  # Create URL
  if (substr(repo, nchar(repo), nchar(repo)) != "/") repo = paste0(repo, "/")
  cran_pkg_loc = paste0(repo, "package=", pkg_name)

  # Establish connection
  suppressWarnings(conn <- try(url(cran_pkg_loc), silent = TRUE))

  # If connection, read in webpage
  if (all(class(conn) != "try-error") ) {
    suppressWarnings(cran_pkg_page <- try(readLines(conn), silent = TRUE))
    close(conn)
  } else {
    return(NULL)
  }

  # Use regex to find version info
  version_line = cran_pkg_page[grep("Version:", cran_pkg_page) + 1]
  version_line = gsub("<(td|\\/td)>","",version_line)
  numeric_version(version_line)

}
